# -*- coding: utf-8 -*-

'''
File Organization Model (FOM)
=============================

A FOM is a JSON file (dictionary) describing how to make filenames from a set of attributes.

FOM definition
--------------

Ex::

    {
        "fom_name": "morphologist-auto-nonoverlap-1.0",

        "fom_import": ["formats-brainvisa-1.0", "brainvisa-formats-3.2.0",
                      "shared-brainvisa-1.0"],

        "attribute_definitions" : {
          "acquisition" : {"default_value" : "default_acquisition"},
          "analysis" : {"default_value" : "default_analysis"},
          "sulci_recognition_session" :  {"default_value" : "default_session"},
          "graph_version": {"default_value": "3.1"},
        },

        "shared_patterns": {
          "acquisition": "<center>/<subject>/t1mri/<acquisition>",
          "analysis": "{acquisition}/<analysis>",
          "recognition_analysis": "{analysis}/folds/<graph_version>/<sulci_recognition_session>_auto",
        },

        "processes" : {
            "Morphologist" : {
                "t1mri":
                    [["input:{acquisition}/<subject>", "images"],
                "imported_t1mri":
                    [["output:{acquisition}/<subject>", "images"]],
                "t1mri_referential":
                    [["output:{acquisition}/registration/RawT1-<subject>_<acquisition>", "Referential"]],
                "reoriented_t1mri":
                    [["output:{acquisition}/<subject>", "images"]],
                "t1mri_nobias":
                    [["output:{analysis}/nobias_<subject>", "images" ]],
                "split_brain":
                    [["output:{analysis}/segmentation/voronoi_<subject>","images"]],
                "left_graph":
                    [["output:{analysis}/folds/<graph_version>/<side><subject>",
                        "Graph and data",
                        {"side": "L", "labelled": "No"}]],
                "left_labelled_graph":
                    [["output:{recognition_analysis}/<side><subject>_<sulci_recognition_session>_auto",
                        "Graph and data", {"side": "L"}]],
                "right_graph":
                    [["output:{analysis}/folds/<graph_version>/<side><subject>",
                        "Graph and data", {"side":"R","labelled":"No"}]],
                "right_labelled_graph":
                    [["output:{recognition_analysis}/<side><subject>_<sulci_recognition_session>_auto",
                        "Graph and data", {"side": "R"}]],
                "Talairach_transform":
                    [["output:{acquisition}/registration/RawT1-<subject>_<acquisition>_TO_Talairach-ACPC",
                        "Transformation matrix"]]
            }
        }
    }

The dictionary may contain:

**fom_name**: string
    identifier of the FOM model. Several FOMs may coexist under different
    identifiers.

**fom_import**: list
    dependencies between FOMs. A Fom may import others to benefit from its
    formats, patterns etc.

**attribute_definitions**: dict
    a dict of predefines attributes. It is generally used to define default
    values for some attributes. Each attribute defined here is also a dict. In
    this sub-dict, the key "default_value" provides the attribute default
    value.

    Attributes don't *need* to be defined here, they are automatically defined
    as they are used in path patterns. But here we can assign them additional
    information (typically default values).

**shared_patterns**: dict
    mapping of reusable patterns. Such patterns will be replaced with their
    contents when used in the file paths patterns. To use such a pattern, place
    it with brackets {} in file paths patterns::

        "input:{acquisition}/<subject>"

    here ``{aquisition}`` is the *shared pattern* "aquisition", and
    ``<subject>`` is the *attribute* "subject"

    A pattern definition may also use attributes between ``<attribute_name>``
    quotes, and / or reuse other patterns between ``{pattern}`` curly brackets.

**processes**: dict
    dictionary of processes supported by the FOM, and path definitions for
    their parameters. The dict is organized by process, then in each process a
    sub-dict contains its parameters definitions. For a given parameter, a list
    of the FOM rules is given: several rules are allowed.

    Process names keys may be a fully qualified module/class name, or a short
    identifier, or a "contextual" name: the name a process has in the context
    of a pipeline.

    A FOM rule is a list of 2 or 3 elements::

        [patterrn, formats, attributes]

    *pattern*: string
        the FOM pattern for the file path of the given parameter. A pattern
        starts with a directory identifier (``input``, ``output``,
        ``shared``), followed by a semicolon (``:``), and a file pattern which
        may contain attributes (``<attrib>``) and/or shared patterns
        (``{pattern}``). Ex::

            "input:{acquisition}/<subject>"

    *formats*: string or list
        name of a format, a formats list, or a list of allowed formats names.
        The formats will rule the possible file extensions.

    *attributes*: dict, optional
        An optional dict assigning locally some attributes values. Ex::

            {"side": "left"}

        The attributes values here are used both to replace attributes values in the current file path pattern, and to select the matching rule when attributes values are given externally (hm, to be checked actually).

A FOM file is a JSON file (actually a JSON/YAML extended file, to allow comments within it), which should be placed in a common directory where the FOM manager will look for it (``share/foms/``)

How to use FOMS
---------------

At higher level, they are used for instance in `CAPSUL <http://brainvisa.info/capsul/>`_.

At lower level, they are used through several classes:

* :class:`FileOrganizationModelManager` manages a set of FOMs, looks for them in a search path, reads them.
* :class:`FileOrganizationModels` represents a FOM rules set.
* :class:`AttributesToPaths` is used to convert a set of attributes into filenames (which is the main use of FOMs).
* :class:`PathToAttributes` performs the reverse operation: match attributes and determines their values from a given filename.

'''

from __future__ import absolute_import
from __future__ import print_function

import sys
import os
import os.path as osp
import stat
import time
import re
import pprint
import sqlite3
import json
import six
from six.moves import range
try:
    import bz2
except ImportError:
    bz2 = None

from collections import OrderedDict


try:
    import yaml

    class json_reader(object):

        '''
        This class has a single static method load that loads an
        JSON file with two features not provided by all JSON readers:

        - JSON syntax is extended. For instance comments are allowed.
        - The order of elements in dictionaries can be preserved by
          using parameter object_pairs_hook=OrderedDict (as in Python
          2.7 JSON reader).
        '''
        @staticmethod
        def load(stream, object_pairs_hook=dict):
            class OrderedLoader(yaml.Loader):
                pass

            def construct_mapping(loader, node):
                loader.flatten_mapping(node)
                return object_pairs_hook(loader.construct_pairs(node))
            OrderedLoader.add_constructor(
                yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
                construct_mapping)
            return yaml.load(stream, OrderedLoader)
except ImportError:
    import json as json_reader

from soma.path import split_path

def deep_update(update, original):
    '''
    Recursively update a dict.
    Subdict's won't be overwritten but also updated.
    '''
    for key, value in six.iteritems(original):
        if not key in update:
            update[key] = value
        elif isinstance(value, dict):
            deep_update(update[key], value)
        elif value != update[key]:
            raise ValueError('In deep_update, for key %s, cannot merge %s and %s' %
                             (repr(key), repr(update[key]), repr(value)))


def read_json(file_name):
    ''' Read a json-like file using yaml or json.
    In case of failure, issue a clearer message with filename, and when
    appropriate a warning about yaml not being installed.
    '''
    try:
        with open(file_name, 'r') as f:
            return json_reader.load(f, object_pairs_hook=OrderedDict)
    except ValueError as e:
        if json_reader.__name__ != 'yaml':
            extra_msg = ' Check your python installation, and perhaps un a "pip install PyYAML" or "easy_install PyYAML"'
        else:
            extra_msg = ''
        raise ValueError('%s: %s. This may be due to yaml module not installed.%s' %
                         (file_name, str(e), extra_msg))


class DirectoryAsDict(object):

    def __new__(cls, directory, cache=None):
        if osp.isdir(directory):
            return super(DirectoryAsDict, cls).__new__(cls, directory, cache)
        else:
            with open(directory) as f:
                return json.load(f)

    def __init__(self, directory, cache=None):
        self.directory = directory
        if cache is None:
            self.cache = DirectoriesCache()
        else:
            self.cache = cache

    def __repr__(self):
        return '<DirectoryAsDict( %s )>' % repr(self.directory)

    def iteritems(self):
        st_content = self.cache.get_directory(self.directory)
        if st_content is not None:
            st, content = st_content
            for i in six.iteritems(content):
                yield i
        else:
            try:
                listdir = os.listdir(self.directory)
            except OSError:
                yield '', [None, None]
                return
            for name in listdir:
                full_path = osp.join(self.directory, name)
                st_content = self.cache.get_directory(full_path)
                if st_content is not None:
                    yield st_content
                else:
                    st = os.stat(full_path)
                    if stat.S_ISDIR(st.st_mode):
                        yield (name, [tuple(st), DirectoryAsDict(full_path)])
                    else:
                        yield (name, [tuple(st), None])

    @staticmethod
    def get_directory(directory, debug=None):
        return DirectoryAsDict._get_directory(
            directory, debug, 0, 0, 0, 0, 0, 0, 0)[0]

    @staticmethod
    def _get_directory(directory, debug, directories, files, links,
                       files_size, path_size, errors, count):
        try:
            listdir = os.listdir(directory)
            result = {}
        except OSError:
            errors += 1
            result = None
        if result is not None:
            for name in listdir:
                if debug and count % 100 == 0:
                    debug.info('%s files=%d, directories=%d, size=%d'
                               % (time.asctime(), files + links, directories,
                                  files_size))
                path_size += len(name)
                count += 1
                full_path = osp.join(directory, name)
                st = os.lstat(full_path)
                if stat.S_ISREG(st.st_mode):
                    files += 1
                    files_size += st.st_size
                    result[name] = [tuple(st), None]
                elif stat.S_ISDIR(st.st_mode):
                    content, directories, files, links, files_size, \
                        path_size, errors, count =  \
                        DirectoryAsDict._get_directory(full_path, debug,
                                                       directories +
                                                       1, files, links, files_size,
                                                       path_size, errors, count)
                    result[name] = [tuple(st), content]
                else:
                    links += 1
                    result[name] = [tuple(st), None]
        return result, directories, files, links, files_size, path_size, \
            errors, count

    @staticmethod
    def paths_to_dict(*paths):
        result = {}
        for path in paths:
            current_dir = result
            path_list = split_path(path)
            for name in path_list[:-1]:
                st_content = current_dir.setdefault(name, [None, {}])
                if st_content[1] is None:
                    st_content[1] = {}
                current_dir = st_content[1]
            current_dir.setdefault(path_list[-1], [None, None])
        return result

    @staticmethod
    def get_statistics(dirdict, debug=None):
        return DirectoryAsDict._get_statistics(
            dirdict, debug, 0, 0, 0, 0, 0, 0, 0)[:-1]

    @staticmethod
    def _get_statistics(dirdict, debug, directories, files, links, files_size,
                        path_size, errors, count):

        if debug and count % 100 == 0:
            debug.info('%s files=%d, directories=%d, size=%d'
                       % (time.asctime(), files + links, directories, files_size))
        count += 1
        for name, content in six.iteritems(dirdict):
            path_size += len(name)
            st, content = content
            if st:
                st = os.stat(st)
                if stat.S_ISREG(st.st_mode):
                    files += 1
                    files_size += st.st_size
                elif stat.S_ISDIR(st.st_mode):
                    if content is None:
                        directories += 1
                        errors += 1
                    else:
                        directories, files, links, files_size, path_size, \
                            errors, count \
                            = DirectoryAsDict._get_statistics(
                                content, debug, directories + 1, files, links,
                                files_size, path_size, errors, count)
                else:
                    links += 1
            else:
                errors += 1
        return (directories, files, links, files_size, path_size, errors,
                count)


class DirectoriesCache(object):

    def __init__(self):
        self.directories = {}

    def add_directory(self, directory, content=None, debug=None):
        if content is None:
            st = tuple(os.stat(directory))
            content = DirectoryAsDict.get_directory(directory, debug=debug)
        else:
            st = None
        self.directories[directory] = [st, content]

    def remove_directory(self, directory):
        del self.directories[directory]

    def has_directory(self, directory):
        return directory in self.directories

    def get_directory(self, directory):
        return self.directories.get(directory)

    def save(self, path):
        if bz2:
            f = bz2.BZ2File(path, 'w')
        else:
            f = open(path, 'w')
        with f:
            json.dump(self.directories, f)

    @classmethod
    def load(cls, path):
        result = cls()
        if bz2:
            try:
                with bz2.BZ2File(path, 'r') as f:
                    result.directories = json.load(f)
            except IOError:
                with open(path, 'r') as f:
                    result.directories = json.load(f)
        else:
            with open(path, 'r') as f:
                result.directories = json.load(f)
        return result


class FileOrganizationModelManager(object):

    '''
    Manage the discovery and instanciation of available FileOrganizationModel
    (FOM). A FOM can be represented as a YAML/JSON file (or a series of
    YAML/JSON files in a directory). This class allows to identify these files
    contained in a predefined set of directories (see find_fom method) and to
    instanciate a FileOrganizationModel for each identified file (see get_fom
    method).
    '''

    def __init__(self, paths=None):
        '''
        Create a FOM manager that will use the given paths to find available FOMs.
        '''
        if paths is None:
            paths = [osp.join(osp.dirname(osp.dirname(osp.dirname(__file__))),
                              'share', 'foms')]
        self.paths = paths
        self._cache = None

    def find_foms(self):
        '''Return a list of file organisation model (FOM) names.
        These FOMs can be loaded with load_foms. FOM files (or directories) are
        looked for in self.paths.'''
        #print('*** find_foms ***')
        #import time
        #t0 = time.time()
        self._cache = {}
        for path in self.paths:
            # print('   ', path)
            if os.path.isdir(path):
                for i in os.listdir(path):
                    full_path = osp.join(path, i)
                    if osp.isdir(full_path):
                        for ext in ('.json', '.yaml'):
                            main_file = osp.join(full_path, i + ext)
                            if osp.exists(main_file):
                                d = read_json(main_file)
                                name = d.get('fom_name')
                                if not name:
                                    raise ValueError(
                                        'file %s does not contain fom_name'
                                        % main_file)
                                self._cache[name] = full_path
                    elif i.endswith('.json') or i.endswith('.yaml'):
                        d = read_json(full_path)
                        if d:
                            name = d.get('fom_name')
                            if not name:
                                raise ValueError(
                                    'file %s does not contain fom_name'
                                    % full_path)
                            self._cache[name] = full_path
        #print('    find_foms done: %f s' % (time.time() - t0))
        return list(self._cache.keys())

    def fom_files(self):
        '''Return a list of file organisation model (FOM) names, as in
        :meth:`find_foms`, but does not clear and reload the cache.
        These FOMs can be loaded with load_foms. FOM files (or directories) are
        looked for in self.paths.'''
        if not self._cache:
            self.find_foms()
        return list(self._cache.keys())

    def clear_cache(self):
        self._cache = None

    def load_foms(self, *names):
        if self._cache is None:
            self.find_foms()
        foms = FileOrganizationModels()
        for name in names:
            foms.import_file(self._cache[name], foms_manager=self)
        return foms

    def file_name(self, fom):
        if self._cache is None:
            self.find_foms()
        return self._cache[fom]

    def read_definition(self, fom_name, done=None):
        jsons = OrderedDict()
        stack = [fom_name]
        while stack:
            fom_name = stack.pop(0)
            if fom_name not in jsons:
                json = jsons[fom_name] = read_json(self.file_name(fom_name))
                stack.extend(json.get('fom_import', []))
        jsons = list(jsons.values())
        result = jsons.pop(0)
        for json in jsons:
            for n in ('attribute_definitions', 'formats', 'format_lists', 'shared_patterns', 'patterns', 'processes'):
                d = json.get(n)
                if d:
                    deep_update(d, result.get(n, {}))
                    result[n] = d
            r = json.get('rules', [])
            if r:
                result.setdefault('rules', []).extend(r)
        return result


class FileOrganizationModels(object):

    def __init__(self):
        self._directories_regex = re.compile(r'{([A-Za-z][A-Za-z0-9_]*)}')
        self._attributes_regex = re.compile('<([^>]+)>')
        self.fom_names = []
        self.attribute_definitions = {
            "fom_name": {
                "descr": "File Organization Model (FOM) in which a pattern is defined.",
                "values": set(self.fom_names),
            },
            "fom_format": {
                "descr": "Format of a file.",
                "values": set(),
            }
        }
        self.formats = {}
        self.format_lists = {}
        self.shared_patterns = {}
        self.patterns = {}
        self.rules = []

    def _expand_shared_pattern(self, pattern):
        expanded_pattern = []
        last_end = 0
        for match in self._directories_regex.finditer(pattern):
            c = pattern[last_end: match.start()]
            if c:
                expanded_pattern.append(c)
            attribute = match.group(1)
            expanded_pattern.append(self.shared_patterns[attribute])
            last_end = match.end()
        if expanded_pattern:
            last = pattern[last_end:]
            if last:
                expanded_pattern.append(last)
            return ''.join(expanded_pattern)
        else:
            return pattern

    def import_file(self, file_or_dict, foms_manager=None):
        if not isinstance(file_or_dict, dict):
            json_dict = read_json(file_or_dict)
        else:
            json_dict = file_or_dict

        foms = json_dict.get('fom_import', [])
        if foms and foms_manager is None:
            raise RuntimeError(
                'Cannot import FOM because no FileOrganizationModelManager has been provided')
        for fom in foms:
            self.import_file(
                foms_manager.file_name(fom), foms_manager=foms_manager)

        fom_name = json_dict['fom_name']
        if fom_name in self.fom_names:
            return
        self.fom_names.append(fom_name)

        # Update attribute definitions
        attribute_definitions = json_dict.get('attribute_definitions')
        if attribute_definitions:
            for attribute, definition in six.iteritems(attribute_definitions):
                existing_definition = self.attribute_definitions.get(
                    attribute)
                values = definition.get('values')
                if existing_definition:
                    existing_values = existing_definition.get('values')
                    # if (existing_values is None) != bool(values is None):
                        # print('ERROR:\nexisting_definition:', existing_definition)
                        # print('definition:', definition)
                        #raise ValueError(
                            #'Incompatible values redefinition for attribute '
                            #'%s, older: %s, new: %s in file: %s' % (
                                #attribute,
                                #existing_values,
                                #values,
                                #file_or_dict))
                    if (definition.get('default_value') is None) \
                            != (existing_definition.get('default_value')
                                is None):
                        raise ValueError(
                            'Incompatible default value redefinition of '
                            'attribute %s, older default: %s, new default: '
                            '%s in file: %s' % (
                                attribute,
                                existing_definition.get('default_value'),
                                definition.get('default_value'), file_or_dict))
                    if values:
                        existing_values.extend(values)
                else:
                    definition = definition.copy()
                    if values is not None:
                        definition['values'] = set(values)
                    self.attribute_definitions[attribute] = definition

        # Process shared patterns to expand the ones that reference other shared
        # patterns
        self.formats.update(json_dict.get('formats', {}))
        self.format_lists.update(json_dict.get('format_lists', {}))
        self.shared_patterns.update(json_dict.get('shared_patterns', {}))
        if self.shared_patterns:
            stack = list(self.shared_patterns.items())
            while stack:
                name, pattern = stack.pop()
                if isinstance(pattern, list):
                    if pattern and isinstance(pattern[0], six.string_types):
                        pattern[0] = self._expand_shared_pattern(pattern[0])
                    else:
                        for i in pattern:
                            i[0] = self._expand_shared_pattern(i[0])
                else:
                    expanded_pattern = self._expand_shared_pattern(pattern)
                    if expanded_pattern != pattern:
                        stack.append((name, expanded_pattern))
                    else:
                        self.shared_patterns[name] = pattern

        rules = json_dict.get('rules')
        patterns = json_dict.get('patterns', {}).copy()
        processes = json_dict.get('processes')

        if rules:
            patterns['fom_dummy'] = rules
        new_patterns = {}
        self._expand_json_patterns(
            patterns, new_patterns, {'fom_name': fom_name})
        self._parse_patterns(new_patterns, self.patterns)

        if processes:
            process_patterns = OrderedDict()
            for process, parameters in six.iteritems(processes):
                process_dict = OrderedDict()
                process_patterns[process] = process_dict
                for parameter, rules in six.iteritems(parameters):
                    if isinstance(rules, six.string_types):
                        rules = self.shared_patterns[rules[1:-1]]
                    parameter_rules = []
                    process_dict[parameter] = parameter_rules
                    for rule in rules:
                        if len(rule) == 2:
                            pattern, formats = rule
                            rule_attributes = {}
                        else:
                            try:
                                pattern, formats, rule_attributes = rule
                            except Exception as e:
                                print('error in FOM: %s, process: %s, param: '
                                    '%s, rule:'
                                    % (fom_name, process, parameter), rule)
                                raise
                        rule_attributes['fom_process'] = process
                        rule_attributes['fom_parameter'] = parameter
                        parameter_rules.append(
                            [pattern, formats, rule_attributes])
            new_patterns = OrderedDict()
            self._expand_json_patterns(
                process_patterns, new_patterns, {'fom_name': fom_name})
            self._parse_patterns(new_patterns, self.patterns)

    def get_attributes_without_value(self):
        att_no_value = {}
        for att in self.shared_patterns:
            if not att.startswith('shared.'):
                for attrec in self._attributes_regex.findall(self.shared_patterns[att]):
                    if attrec not in att_no_value:
                        att_no_value[attrec] = ''

        return att_no_value

    def selected_rules(self, selection, debug=None):
        if selection:
            format = selection.get('format')
            for rule_pattern, rule_attributes in self.rules:
                if debug:
                    debug.debug('selected_rules: %s, %s' %
                                (repr(rule_pattern), repr(rule_attributes)))
                rule_formats = rule_attributes.get('fom_formats', [])
                if format:
                    if format in ('fom_first', 'fom_preferred'):
                        if not rule_formats:
                            if debug:
                                debug.debug(
                                    'selected_rules: -- no format in rule')
                            continue
                    elif format not in rule_formats:
                        if debug:
                            debug.debug('selected_rules: -- format %s not in %s' % (repr(
                                format), repr(rule_formats)))
                        continue
                keep = True
                for attribute, selection_value in six.iteritems(selection):
                    if attribute == 'format':
                        continue
                    rule_value = rule_attributes.get(attribute)
                    if rule_value is None or rule_value != selection_value:
                        if debug:
                            debug.debug('selected_rules: -- selection value %s != rule value %s' %
                                        (repr(selection_value), repr(rule_value)))
                        keep = False
                        break
                if keep:
                    if debug:
                        debug.debug('selected_rules: ++')
                    yield (rule_pattern, rule_attributes)
        else:
            for rule in self.rules:
                yield rule

    def _expand_json_patterns(self, json_patterns, parent, parent_attributes):
        attributes = parent_attributes.copy()
        attributes.update(json_patterns.get('fom_attributes', {}))
        for attribute, value in six.iteritems(attributes):
            if attribute not in self.attribute_definitions:
                self.attribute_definitions[
                    attribute] = {'values': set((value,))}
            else:
                values = self.attribute_definitions[
                    attribute].setdefault('values', set())
                values.add(value)

        key_attribute = json_patterns.get('fom_key_attribute', None)
        if key_attribute:
            self.attribute_definitions.setdefault(key_attribute, {})
            # raise ValueError( 'Attribute "%s" must be declared in
            # attribute_definitions' % key_attribute )

        for key, value in six.iteritems(json_patterns):
            if key.startswith('fom_') and key != 'fom_dummy':
                continue
            if key_attribute:
                attributes[key_attribute] = key
                self.attribute_definitions[key_attribute].setdefault(
                    'values', set()).add(key)
            if isinstance(value, dict):
                self._expand_json_patterns(
                    value, parent.setdefault(key, OrderedDict()), attributes)
            else:
                rules = []
                parent[key] = rules
                for rule in value:
                    if len(rule) == 2:
                        pattern, format_list = rule
                        rule_attributes = attributes.copy()
                    else:
                        pattern, format_list, rule_attributes = rule
                        for attribute, value in six.iteritems(rule_attributes):
                            definition = self.attribute_definitions.setdefault(
                                attribute, {})
                            values = definition.setdefault('values', set())
                            values.add(value)
                        if attributes:
                            new_attributes = attributes.copy()
                            new_attributes.update(rule_attributes)
                            rule_attributes = new_attributes

                    # Expand format_list
                    rule_formats = []
                    if isinstance(format_list, six.string_types):
                        format_list = [format_list]
                    if format_list:
                        for format in format_list:
                            formats = self.format_lists.get(format)
                            if formats is not None:
                                for f in formats:
                                    rule_formats.append(f)
                            else:
                                rule_formats.append(format)
                        rule_attributes['fom_formats'] = rule_formats

                    # Expand patterns in rules
                    while True:
                        expanded_pattern = []
                        last_end = 0
                        for match in self._directories_regex.finditer(pattern):
                            c = pattern[last_end: match.start()]
                            if c:
                                expanded_pattern.append(c)
                            attribute = match.group(1)
                            expanded_pattern.append(
                                self.shared_patterns[attribute])
                            last_end = match.end()
                        if expanded_pattern:
                            last = pattern[last_end:]
                            if last:
                                expanded_pattern.append(last)
                            pattern = ''.join(expanded_pattern)
                        else:
                            break
                    rules.append([pattern, rule_attributes])

    def _parse_patterns(self, patterns, dest_patterns):
        for key, value in six.iteritems(patterns):
            if isinstance(value, dict):
                self._parse_patterns(
                    value, dest_patterns.setdefault(key, OrderedDict()))
            else:
                pattern_rules = dest_patterns.setdefault(key, [])
                for rule in value:
                    pattern, rule_attributes = rule
                    for attribute in self._attributes_regex.findall(pattern):
                        s = attribute.find('|')
                        if s > 0:
                            attribute = attribute[:s]
                        definition = self.attribute_definitions.setdefault(
                            attribute, {})
                        value = rule_attributes.get(attribute)
                        if value is not None:
                            definition.setdefault('values', set()).add(value)
                        elif 'fom_open_value' not in definition:
                            definition['fom_open_value'] = True
                            # raise ValueError( 'Attribute "%s" must be
                            # declared in attribute_definitions' % attribute )
                        if attribute in rule_attributes:
                            pattern = pattern.replace(
                                '<' + attribute + '>', rule_attributes[attribute])
                    i = pattern.find(':')
                    if i > 0:
                        rule_attributes['fom_directory'] = pattern[:i]
                        pattern = pattern[i + 1:]
                    pattern_rules.append([pattern, rule_attributes])
                    self.rules.append([pattern, rule_attributes])

    def pprint(self, out=sys.stdout):
        for i in ('fom_names', 'attribute_definitions', 'formats', 'format_lists', 'shared_patterns', 'patterns', 'rules'):
            print('-' * 20, i, '-' * 20, file=out)
            pprint.pprint(getattr(self, i), out)


class PathToAttributes(object):

    '''
    Utility class for file paths -> attributes set transformation.
    Part of the FOM engine.
    '''

    def __init__(self, foms, selection=None):
        self._attributes_regex = re.compile('<([^>]+)>')
        self.hierarchical_patterns = OrderedDict()
        for rule_pattern, rule_attributes in foms.selected_rules(selection):
            rule_formats = rule_attributes.get('fom_formats', [])
            parent = self.hierarchical_patterns
            attributes_found = set()
            splited_pattern = rule_pattern.split('/')
            count = 0
            for pattern in splited_pattern:
                count += 1
                regex = ['^']
                last_end = 0
                for match in self._attributes_regex.finditer(pattern):
                    c = pattern[last_end: match.start()]
                    if c:
                        regex.append(re.escape(c))
                    attribute = match.group(1)
                    s = attribute.find('|')
                    if s > 0:
                        attribute_re = attribute[s + 1:]
                        attribute = attribute[:s]
                    else:
                        attribute_re = '[^/]*'
                    if attribute in attributes_found:
                        regex.append('%(' + attribute + ')s')
                    else:
                        attribute_type = foms.attribute_definitions[attribute]
                        values = attribute_type.get('values')
                        if values and not attribute_type.get('fom_open_value', True):
                            regex.append(
                                '(?P<%s>%s)' % (attribute, '|'.join('(?:' + re.escape(i) + ')' for i in values)))
                        else:
                            regex.append(
                                '(?P<%s>%s)' % (attribute, attribute_re))
                        attributes_found.add(attribute)
                    last_end = match.end()
                last = pattern[last_end:]
                if last:
                    regex.append(re.escape(last))
                if count == len(splited_pattern):
                    if rule_formats:
                        for format in rule_formats:
                            if format not in foms.formats:
                                print('format "%s" not if FOM "%s"'
                                      % (format, foms.fom_names))
                            extension = foms.formats[format]
                            d = rule_attributes.copy()
                            d['fom_format'] = format
                            d.pop('fom_formats', None)
                            parent.setdefault(''.join(regex) + '$', [OrderedDict(), OrderedDict()])[
                                0].setdefault(extension, []).append(d)
                    else:
                        parent.setdefault(''.join(regex) + '$', [OrderedDict(), OrderedDict()])[
                            0].setdefault('', []).append(rule_attributes)
                else:
                    parent = parent.setdefault(
                        ''.join(regex) + '$', [OrderedDict(), OrderedDict()])[1]

    def pprint(self, file=sys.stdout):
        self._pprint(file, self.hierarchical_patterns, 0)

    def _pprint(self, file, node, indent):
        if node:
            print('  ' * indent + '{', file=file)
            for pattern, rules_subpattern in six.iteritems(node):
                ext_rules, subpattern = rules_subpattern
                print('  ' * (indent + 1) + repr(pattern) + ': { (', file=file)
                if ext_rules:
                    print('  ' * (indent + 1) + '{', file=file)
                    for ext, rules in six.iteritems(ext_rules):
                        print('  ' * \
                            (indent + 2) + repr(ext) + ': ', repr(rules),
                          file=file)
                    print('  ' * (indent + 1) + '},', file=file)
                else:
                    print('  ' * (indent + 1) + '{},', file=file)
                self._pprint(file, subpattern, indent + 1)
                print('),', file=file)
            print('  ' * indent + '}', file=file, end=' ')
        else:
            print('  ' * indent + '{}', file=file, end=' ')

    def parse_directory(self, dirdict, single_match=False, all_unknown=False, log=None):
        if isinstance(dirdict, six.string_types):
            dirdict = DirectoryAsDict.paths_to_dict(dirdict)
        return self._parse_directory(dirdict, [([], self.hierarchical_patterns, {})], single_match, all_unknown, log)

    def _parse_directory(self, dirdict, parsing_list, single_match, all_unknown, log):
        for name, content in six.iteritems(dirdict):
            st, content = content
            # Split extention on left most dot
            l = name.split('.')
            possible_extension_split = [('.'.join(l[:i]), '.'.join(l[i:]))
                                        for i in range(1, len(l) + 1)]
            # split = name.split('.', 1)
            # name_no_ext = split[0]
            # if len(split) == 2:
                # ext = split[1]
            # else:
                # ext = ''

            matched_directories = []
            matched = False
            sent = False
            recurse_parsing_list = []
            for path, hierarchical_patterns, pattern_attributes in parsing_list:
                if log:
                    log.debug('?? ' + name + ' ' + repr(
                        pattern_attributes) + ' ' + repr(list(hierarchical_patterns.keys())))
                branch_matched = False
                for pattern, rules_subpattern \
                        in six.iteritems(hierarchical_patterns):
                    stop_parsing = False
                    for name_no_ext, ext in possible_extension_split:
                        ext_rules, subpattern = rules_subpattern
                        pattern = pattern % pattern_attributes
                        match = re.match(pattern, name_no_ext)
                        if log:
                            log.debug(
                                'try %s for %s' % (repr(pattern), repr(name_no_ext)))
                        if match:
                            if log:
                                log.debug('match ' + pattern)
                            new_attributes = match.groupdict()
                            new_attributes.update(pattern_attributes)

                            rules = ext_rules.get(ext)
                            if (subpattern and
                                    not ext and
                                    (st is None or
                                     stat.S_ISDIR(st[0])) and
                                    content is not None):
                                matched = branch_matched = True
                                stop_parsing = single_match
                                full_path = path + [name]
                                if log:
                                    log.debug('directory matched: %s %s' % (
                                        repr(full_path), (repr([i[0] for i in six.iteritems(content)]) if content else None)))
                                matched_directories.append(
                                    (full_path, subpattern, new_attributes))
                            else:
                                if log:
                                    log.debug(
                                        'no directory matched for %s' % repr(name))
                            if rules is not None and ext:
                                matched = branch_matched = True
                                if log:
                                    log.debug(
                                        'extension matched: ' + repr(ext))
                                for rule_attributes in rules:
                                    yield_attributes = new_attributes.copy()
                                    yield_attributes.update(rule_attributes)
                                    stop_parsing = single_match or yield_attributes.pop(
                                        'fom_stop_parsing', False)
                                    if log:
                                        log.debug(
                                            '-> ' + '/'.join(path + [name]) + ' ' + repr(yield_attributes))
                                    sent = True
                                    yield path + [name], st, yield_attributes
                                break
                            else:
                                if log:
                                    log.debug(
                                        'no extension matched: ' + repr(ext))
                        if stop_parsing:
                            break
                    if stop_parsing:
                        break
                if branch_matched:
                    for full_path, subpattern, new_attributes in matched_directories:
                        if content:
                            recurse_parsing_list.append(
                                (full_path, subpattern, new_attributes))
            if recurse_parsing_list:
                for i in self._parse_directory(content, recurse_parsing_list, single_match, all_unknown, log):
                    yield i
            if not matched and all_unknown:
                if log:
                    log.debug('-> ' + '/'.join(path + [name]) + ' None')
                sent = True
                yield path + [name], st, None
                if content:
                    for i in self._parse_unknown_directory(content, path + [name], log):
                        yield i
            if not sent and all_unknown:
                if log:
                    log.debug('-> ' + '/'.join(path + [name]) + ' None')
                yield path + [name], st, None

    def _parse_unknown_directory(self, dirdict, path, log):
        for name, content in six.iteritems(dirdict):
            st, content = content
            if log:
                log.debug('?-> ' + '/'.join(path + [name]) + ' None')
            yield path + [name], st, None
            if content is not None:
                for i in self._parse_unknown_directory(content, path + [name], log):
                    yield i

    def parse_path(self, path, single_match=False, log=None):
        dirdict = DirectoryAsDict.paths_to_dict(path)
        spath = split_path(path)
        for p, s, a in self.parse_directory(dirdict, single_match=single_match, log=log):
            if spath == p:
                yield (p, s, a)


class AttributesToPaths(object):

    '''
    Utility class for attributes set -> file paths transformation.
    Part of the FOM engine.
    '''

    def __init__(self, foms, selection=None, directories={}, preferred_formats=set(), debug=None):
        self.foms = foms
        self.selection = selection or {}
        self.directories = directories
        self._db = sqlite3.connect(':memory:', check_same_thread=False)
        self._db.execute('PRAGMA journal_mode = OFF;')
        self._db.execute('PRAGMA synchronous = OFF;')
        self.all_attributes = tuple(
            i for i in self.foms.attribute_definitions if i != 'fom_formats')
        self.default_values = dict(
            (i, self.foms.attribute_definitions[i]['default_value']) for i in self.all_attributes if 'default_value' in self.foms.attribute_definitions[i])
        self.non_discriminant_attributes = set(
            i for i in self.all_attributes if not self.foms.attribute_definitions[i].get('discriminant', True))
        fom_format_index = self.all_attributes.index('fom_format')
        sql = 'CREATE TABLE rules ( %s, _fom_first, _fom_preferred_format, _fom_rule )' % ','.join(repr('_' + str(i))
                                for i in self.all_attributes)
        if debug:
            debug.debug(sql)
        self._db.execute(sql)
        columns = ['_%s' %
                   i for i in self.all_attributes + ('fom_first', 'fom_preferred_format')]
        sql = 'CREATE INDEX rules_index ON rules (%s)' % ','.join(columns)
        self._db.execute(sql)
        for i in columns:
            sql = 'CREATE INDEX rules%s_index ON rules (%s)' % (i, i)
            self._db.execute(sql)
        sql_insert = 'INSERT INTO rules VALUES ( %s )' % ','.join(
            '?' for i in range(len(self.all_attributes) + 3))
        self.rules = []
        for pattern, rule_attributes in foms.selected_rules(self.selection, debug=debug):
            if debug:
                debug.debug(
                    'pattern: ' + pattern + ' ' + repr(rule_attributes))
            pattern_attributes = set((i if '|' not in i else i[: i.find('|')])
                                     for i in self.foms._attributes_regex.findall(pattern))
            values = []
            for attribute in self.all_attributes:
                value = rule_attributes.get(attribute)
                if not value and attribute in pattern_attributes:
                    value = ''
                values.append(value)
            values.append(True)
            values.append(False)
            values.append(len(self.rules))
            self.rules.append(
                (re.sub(r'<([^>|]*)(\|[^>]*)?>', r'%(\1)s', pattern), rule_attributes))
            fom_formats = rule_attributes.get('fom_formats')
            if fom_formats and 'fom_format' not in rule_attributes:
                first = True
                for format in fom_formats:
                    if format in preferred_formats:
                        preferred_format = format
                        break
                else:
                    preferred_format = fom_formats[0]
                sys.stdout.flush()
                for format in fom_formats:
                    values[fom_format_index] = format
                    values[-3] = first
                    values[-2] = bool(format == preferred_format)
                    first = False
                    if debug:
                        debug.debug(sql_insert + ' ' + repr(values))
                    self._db.execute(sql_insert, values)
            else:
                if debug:
                    debug.debug(sql_insert + ' ' + repr(values))
                self._db.execute(sql_insert, values)
        self._db.commit()

    def find_paths(self, attributes={}, debug=None):
        if debug:
            debug.debug('!find_path! %s' % repr(attributes))
        d = self.selection.copy()
        d.update(attributes)
        attributes = d
        select = []
        values = []
        selection_attributes = {}
        default_values = []
        for attribute in self.all_attributes:
            value = attributes.get(attribute)
            if value is None:
                value = self.selection.get(attribute)
            if value is None:
                default_value = self.default_values.get(attribute)
                if default_value is not None:
                    default_values.append((attribute, default_value))
                    if attribute not in self.non_discriminant_attributes:
                        select.append(
                            '(_' + attribute + " IN ('','%s') OR _" % default_value + attribute + ' IS NULL )')
                else:
                    if attribute not in self.non_discriminant_attributes:
                        select.append(
                            '(_' + attribute + " != '' OR _" + attribute + ' IS NULL )')
            elif attribute == 'fom_format':
                selected_format = attributes.get('fom_format')
                if selected_format == 'fom_first':
                    select.append('_fom_first = 1')
                elif selected_format == 'fom_preferred':
                    select.append('_fom_preferred_format = 1')
                elif isinstance(value, list):
                    select.append('_' + attribute + " IN (%s)" %
                                  ','.join('?' for i in value))
                    values.extend(value)
                else:
                    select.append('_' + attribute + " = ?")
                    values.append(value)
            elif isinstance(value, list):
                if attribute not in self.non_discriminant_attributes:
                    select.append('_' + attribute + " IN ( %s, '' )" %
                                  ','.join('?' for i in value))
                    values.extend(value)
            else:
                if attribute not in self.non_discriminant_attributes:
                    select.append('_' + attribute + " IN ( ?, '' )")
                    values.append(value)
                    selection_attributes[attribute] = value
        columns = ['_fom_rule', '_fom_format'] + ['_' + i[0]
                                                  for i in default_values]
        sql = 'SELECT %s FROM rules WHERE %s' % (','.join(columns), ' AND '.join(
            select))
        if debug:
            debug.debug('!sql! %s' %
                        (sql.replace('?', '%s') % tuple(repr(i) for i in values)))
        for row in self._db.execute(sql, values):
            rule_index, format = row[:2]
            row = row[2:]
            # bool_output = False
            rule, rule_attributes = self.rules[rule_index]
            rule_attributes = rule_attributes.copy()
            default_attributes = {}
            for i in range(len(default_values)):
                if not row[i]:
                    rule_attributes[
                        default_values[i][0]] = default_values[i][1]
                    default_attributes[
                        default_values[i][0]] = default_values[i][1]
            # rule_attributes = self.foms.rules[ rule_index ][ 1 ].copy()
            fom_formats = rule_attributes.pop('fom_formats', [])

            # if rule_attributes.get( 'fom_directory' ) == 'output':
                # bool_output=True

            if debug:
                debug.debug('!rule matching! %s' %
                            repr((rule, fom_formats, rule_attributes)))
            if format:
                ext = self.foms.formats[format]
                if ext != '':
                    ext = '.' + ext
                rule_attributes['fom_format'] = format
                default_attributes.update(attributes)
                try:
                    path = rule % default_attributes + ext
                except KeyError:
                    continue
                if debug:
                    debug.debug('!single format! %s: %s' % (
                        format, path))
                r = self._join_directory(
                    path, rule_attributes,
                    selection_attributes)
                if r:
                    if debug:
                        debug.debug('!-->! %s' % repr(r))
                    yield r
            else:
                if fom_formats:
                    for f in fom_formats:
                        ext = self.foms.formats[f]
                        if ext != '':
                            ext = '.' + ext
                        rule_attributes['fom_format'] = f
                        default_attributes.update(attributes)
                        try:
                            path = rule % default_attributes + ext
                        except KeyError:
                            continue
                        if debug:
                            debug.debug('!format from fom_formats! %s: %s' %
                                        (f, path))
                        r = self._join_directory(
                            path, rule_attributes,
                            selection_attributes)
                        if r:
                            if debug:
                                debug.debug('!-->! %s' % repr(r))
                            yield r
                else:
                    default_attributes.update(attributes)
                    try:
                        path = rule % default_attributes
                    except KeyError:
                        continue
                    if debug:
                        debug.debug('!no format! %s' % path)
                    r = self._join_directory(
                        path, rule_attributes,
                        selection_attributes)
                    if r:
                        if debug:
                            debug.debug('!-->! %s' % repr(r))
                        yield r

    def find_discriminant_attributes(self, **selection):
        result = []
        if self.rules:
            for attribute in self.all_attributes:
                sql = 'SELECT DISTINCT "%s" FROM rules' % ('_' + attribute)
                if selection:
                    sql += ' WHERE ' + \
                        ' AND '.join('_' + i + ' = ?' for i in selection)
                    values = list(self._db.execute(sql,
                                                   list(selection.values())))
                else:
                    values = list(self._db.execute(sql))
                if values and (len(values) > 1 or ('',) in values):
                    result.append(attribute)
        return result

    def find_attributes_values(self, **selection):
        result = {}
        if self.rules:
            for attribute in self.all_attributes:
                sql = 'SELECT DISTINCT "%s" FROM rules' % ('_' + attribute)
                if selection:
                    sql += ' WHERE ' + \
                        ' AND '.join('_' + i + ' = ?' for i in selection)
                    values = list(self._db.execute(sql,
                                                   list(selection.values())))
                else:
                    values = list(self._db.execute(sql))
                result[attribute] = values
        return result

    def _join_directory(self, path, rule_attributes, selection_attributes):
        attributes = selection_attributes.copy()
        attributes.update(rule_attributes)
        fom_directory = rule_attributes.get('fom_directory')
        if fom_directory:
            directory = self.directories.get(fom_directory)
            if directory:
                return (osp.join(directory, *path.split('/')), attributes)
                #return (osp.join(directory, path), attributes)
        return (osp.join(*path.split('/')), attributes)
        #return (path, attributes)

    def allowed_formats_for_parameter(self, process_name, param):
        formats = []
        for rule, attributes in self.rules:
            if attributes.get('fom_process') == process_name \
                    and attributes.get('fom_parameter') == param:
                rformats = attributes.get('fom_formats', [])
                for format in rformats:
                    if format not in formats:
                        formats.append(format)
        return formats

    def allowed_extensions_for_parameter(self, **kwargs):
        '''
        Either formats or (process_name and param) should be passed
        kwargs
        ------
        formats: list of str
            if provided only this parameter will be used
        process_name: str
            name of the process
        param: str
            name of the process parameter
        '''
        if 'formats' in kwargs:
            formats = list(kwargs['formats'])
        elif 'process_name' in kwargs and 'param' in kwargs:
            process_name = kwargs['process_name']
            param = kwargs['param']
            formats = self.allowed_formats_for_parameter(process_name, param)
        else:
            raise KeyError('Either formats or (process_name and param) should '
                           'be passed')
        exts = set()
        while formats:
            format = formats.pop(0)
            if format not in self.foms.formats \
                    and format in self.foms.format_lists:
                formats += self.foms.format_lists[format]
                continue
            exts.add(self.foms.formats[format])

        return sorted(exts)



def call_before_application_initialization(application):
    try:
        from traits.api import ListStr
    except ImportError:
        from enthought.traits.api import ListStr

    application.add_trait(
        'fom_path',
        ListStr(descr='Path for finding file organization models'))
    if application.install_directory:
        application.fom_path = [osp.join(application.install_directory,
                                         'share', 'foms')]


def call_after_application_initialization(application):
    application.fom_manager = FileOrganizationModelManager(
        application.fom_path)


if __name__ == '__main__':
    from soma.application import Application
    # First thing to do is to create an Application with name and version
    app = Application('soma.fom', '1.0')
    # Register module to load and call functions before and/or after
    # initialization
    app.plugin_modules.append('soma.fom')
    # Application initialization (e.g. configuration file may be read here)
    app.initialize()
    # process_completion( sys.argv[1], {'spm' : '/here/is/spm', 'shared' :
    # '/volatile/bouin/build/trunk/share/brainvisa-share-4.4' })

    import logging
    # logging.root.setLevel( logging.DEBUG )
    fom = app.fom_manager.load_foms('morphologist-brainvisa-pipeline-1.0')
    # atp = AttributesToPaths( fom, selection={ 'fom_process':'morphologistSimp.SimplifiedMorphologist' },
                             # preferred_formats=set( ('NIFTI',) ),
                             # debug=logging )
    # form='MINC'
    # form=','+'MESH'
    # print('form',form)
    directories = {"input_directory": "/input",
                   "output_directory": "/output",
                   "shared_directory": "/shared"}
    fomr = ['NIFTI', 'MESH']
    atp = AttributesToPaths(
        fom, selection={'fom_process': 'morphologistPipeline.HeadMesh'},
        preferred_formats=fomr, directories=directories,
        debug=logging)
    d = {
        'protocol': u'subjects', 'analysis': 'default_analysis', 'fom_parameter': 'head_mesh',
        'acquisition': 'default_acquisition', 'subject': u'002_S_0816_S18402_I40732', 'fom_format': 'fom_preferred'}
    for p, a in atp.find_paths(d, debug=logging):
        print('->', repr(p), a)
    # for parameter in fom.patterns[ 'morphologistSimp.SimplifiedMorphologist' ]:
        # print('- %s' % parameter)
        # for p, a in atp.find_paths( { 'fom_parameter': parameter,
                                      #'protocol': 'c',
                                      #'subject': 's',
                                      #'analysis': 'p',
                                      #'acquisition': 'a',
                                      #'fom_format': 'fom_preferred',
                                #} ):
        # print(' ', repr( p ), a)



# -*- coding: utf-8 -*-
 
from __future__ import print_function

from __future__ import absolute_import
import unittest
import shutil
import os
import tempfile
from soma.sorted_dictionary import SortedDictionary
import pickle


class TestSortedDictionary(unittest.TestCase):

    def test_sorted_dictionary(self):
        d0 = SortedDictionary()
        self.assertEqual(dict(d0), {})

        d1 = SortedDictionary(
            ('titi', {'bubu': '50', 'turlute': 12}),
            ('toto', 'val"u\'e'),
            ('tutu', [0, 1, 2, [u'papa', 5]]))
        d2 = SortedDictionary(
            ('tutu', [0, 1, 2, [u'papa', 5]]),
            ('toto', 'val"u\'e'),
            ('titi', {'bubu': '50', 'turlute': 12}))

        self.assertEqual(dict(d1), dict(d2))
        self.assertNotEqual(list(d1.keys()), list(d2.keys()))
        self.assertEqual(d1, dict(d1))
        self.assertEqual(d1, SortedDictionary(list(dict(d1).items())))

        d1['titi'] = 'babar'
        d2['titi'] = 'bubur'
        self.assertNotEqual(dict(d1), dict(d2))
        d2['titi'] = 'babar'
        self.assertEqual(dict(d1), dict(d2))
        d1['ababo'] = 43.65
        self.assertEqual(list(d1.keys()), ['titi', 'toto', 'tutu', 'ababo'])

        del d1['titi']
        del d1['ababo']
        del d2['titi']
        self.assertEqual(dict(d1), dict(d2))
        self.assertEqual(list(d2.keys()), ['tutu', 'toto'])
        p = pickle.dumps(d1)
        p2 = pickle.loads(p)
        self.assertTrue(isinstance(p2, SortedDictionary))
        self.assertEqual(d1, p2)

        d1.insert(1, 'babar', 'new item')
        self.assertEqual(list(d1.keys()), ['toto', 'babar', 'tutu'])
        self.assertRaises(KeyError, d1.insert, 2, 'babar', 'other')
        self.assertEqual(d1.index('babar'), 1)


def test():
    suite = unittest.TestLoader().loadTestsFromTestCase(TestSortedDictionary)
    runtime = unittest.TextTestRunner(verbosity=2).run(suite)
    return runtime.wasSuccessful()


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import sys
import six


'''
Framework to serialize/deserialize Python objects in JSON. The main
feature of this framework is that deserialization of a JSON value requires
no other information than the JSON itself. It is not necessary to know the
class of the object to deserialize. This imposes that the serialization
contains a reference to a factory that is responsible to build the Python
object given its JSON.
'''

class JSONSerializable(object):
    '''
    Instances of classes deriving from `JSONSerializable` can be serialized
    in a JSON compatible object with :meth:`to_json` method. This JSON object
    contains a reference to a factory as well as the full state of the 
    instance. Calling the factory with the state of the instance makes it
    possible to recreate the instance (this is what function :func:`from_json`
    is doing). A typical usage of this serialization system is to store the
    JSON serialization of an instance in a configuration file (or database)
    and latter (possibly in another Python instance) recreate the same
    instance. This is an alternative to pickle allowing to use a standard
    representation format (JSON) and to have full control on instances
    creation.
    '''
    def to_json(self):
        '''
        Return a JSON serialization of self. The returned object can be given
        to :func:`from_json` to create another instance that is eqivalent to
        self.
        Here, equivalent means that all attributes values are the same and the
        methods called with the same parameters give the same results.

        See :func:`from_json` to have insight of what is a JSON serialization
        object.
        '''
        raise NotImplementedError()

def to_json(object):
    '''
    Return a JSON serialization of an object.
    Standard types (None, bool, int, float, str, etc.) are using
    json.dumps whereas other type of object must define a to_json
    method returning the serialization.
    '''
    serializer = getattr(object, 'to_json', None)
    if serializer is None:
        if object is None:
            return None
        raise ValueError('Cannot serialize in JSON object of type {0}'.format(type(object)))
    else:
        return serializer()

def from_json(json_serialization):
    '''
    Takes a JSON serialization object (typically created by a
    :meth:`JSONSerializable.to_json` method) and create and return the
    corresponding instance.
    
    A JSON serialization object can have one of the following structures:
    - A simple string containing a factory reference
    - A list with one, two or three of the following items:

        - factory: a mandatory item containing a reference to a factory
          (see :func:`find_factory`)
        - args: an optional item containing a list of parameters values
          for the factory
        - kwargs: an optional item containing a dictionary of parameters
          for the factory
    
    A reference to a factory identifies a :class:`callable` (e.g a function or
    a class) in a Python module. It is a string containing the module name
    and the callable name separated by a dot. For instance 
    ``'catidb.data_models.catidb_3_4'`` would identify the *catidb_3_4*
    callable in the *catidb.data_models* module.
    '''
    if json_serialization is None:
        return None
    elif isinstance(json_serialization, six.string_types):
        callable = find_factory(json_serialization)
        return callable()
    elif isinstance(json_serialization, list):
        if len(json_serialization) == 3:
            callable, args, kwargs = json_serialization
            callable = find_factory(callable)
            return callable(*args, **kwargs)
        elif len(json_serialization) == 2:
            callable, args_or_kwargs = json_serialization
            if isinstance(args_or_kwargs, list):
                callable = find_factory(callable)
                return callable(*args_or_kwargs)
            elif isinstance(args_or_kwargs, dict):
                callable = find_factory(callable)
                return callable(**args_or_kwargs)
        elif len(json_serialization) == 1:
            callable = find_factory(json_serialization[0])
            return callable()
    raise ValueError('Object is not a valid JSON serialization: %s' % repr(json_serialization))


def find_factory(reference):
    '''
    Finds a factory callable given its reference. The reference is simply
    a string containing a module name and the name of fatory in that module
    separated by a dot. For instance ``'my_packages.my_module.my_factory'``
    would refer to the item *my_factory* in the module
    ``my_packages.my_module``.
    This funcion simply loads the module and returns the module attribute with
    the factory name.
    
    :class:`exceptions.ValueError` is raised if the factory cannot be found.
    '''
    split = reference.rsplit('.', 1)
    if len(split) != 2:
        raise ValueError('%s is not a valid reference to a factory' % reference)
    module_name, item_name = split
    try:
        __import__(module_name)
    except ImportError as e:
        raise ValueError('{0} is not a valid reference to a factory: {1}'.format(reference, e))
    module = sys.modules[module_name]
    factory = getattr(module, item_name, None)
    if factory is None:
        raise ValueError('{0} is not a valid reference to a factory: module {1} has no attribute {2}'.format(reference, module_name, item_name))
    return factory



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
A minf tree is used to convert Python objects into a structure that can be written in any minf format. When a Python object is written into a minf file, if it cannot be directly stored in the choosen minf format, it is transformed in a minf tree by a L{MinfReducer}. During reading, minf trees are converted into Python objects by a L{MinfExpander}. Whatever the minf format used (XML, Python, HDF5, etc.) reading and writing objects is always done with a L{MinfReducer}/L{MinfExpander} pair. Each L{MinfReducer}/L{MinfExpander} pair is identified by a name. The name of one L{MinfReducer}/L{MinfExpander} pair must be choosen when writing a minf file, this name is recorded in the minf file and used for reading.

A minf tree is always accessed via an iterator on its content. This content is composed of atoms and special structure objects. Atoms are any Python objects that can be directly stored in the minf file format (without need for reducing or expanding these objects). Other objects are reduced in a structure starting with a L{StartStructure} instance and terminated by an L{EndStructure} instance. Between the L{StartStructure}/L{EndStructure} pair, there can be any atoms and/or special structure objects.

* author: Yann Cointepas
* organization: `NeuroSpin <http://www.neurospin.org>`_
* license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
'''
from __future__ import absolute_import
from six.moves import range
__docformat__ = "restructuredtext en"

import types
import sys
import six
from soma.translation import translate as _
from soma.undefined import Undefined
from soma.minf.error import MinfError
try:
    from soma.signature.api import HasSignature, Sequence
except ImportError:
    class HasSignature(object):
        pass
    class Sequence(object):
        pass

#: Type name of a minf structure.
#: @see: L{StartStructure}
minfStructure = 'minf'
#: Type name of a list structure.
#: @see: L{StartStructure}
listStructure = 'list'
#: Type name of a dictionary structure.
#: @see: L{StartStructure}
dictStructure = 'dict'


#------------------------------------------------------------------------------
class StartStructure(object):

    '''
    When iterating over a minf tree, a L{StartStructure} indicate the begining of
    a subtree. Subtrees are identified by their type which is a string
    defining how the object has been reduced (and how to expand it). There are
    three built-in structures:
      - L{minfStructure}: this is the first element of a minf tree. The content
        of this tree is a series of objects (i.e. atoms) or reduced objects.
        The L{StartStructure} instances of type L{minfStructure} have a mandatory
        'reduction' attribute containing the name of the L{MinfReducer}/L{MinfExpander} pair
        corresponding to this minf file.
      - L{listStructure}: the content of the subtree is a series of objects or
        reduced objects. By default, a L{listStructure} is expanded as a Python
        list.
        The L{StartStructure} instances of type L{listStructure} accepts an
        optional 'length' attribute containing the number of elements in the
        list.
      - dictStructure: the content of the subtree is a series of pair of objects or
        reduced objects. In each pair, the first item represent a key and the
        second item represent a value. By default, a L{dictStructure} is expanded
        as a Python dictionary.
        The L{StartStructure} instances of type L{dictStructure} accepts an
        optional 'length' attribute containing the number of key/value pairs in
        the dictionary.

    Any non-builtin structure type must be registered in the L{MinfReducer}/L{MinfExpander}
    defined in the main L{minfStructure} tree.

    '''

    def __init__(self, structureNodeType, identifier=None, **attributes):
        '''
        @param structureNodeType: type of the structure (see L{StartStructure}).
        @type  structureNodeType: string
        @param identifier: EXPERIMENTAL: identifier of the subtree (used for
          referencing).
        @type  identifier: string or C{None}
        @param attributes: attributes of the structure. Possible name/values
          for attributes are dependant on the structure type.
        @type  attributes: dict
        '''
        self.type = structureNodeType
        self.identifier = identifier
        self.attributes = attributes

    def __repr__(self):
        if self.identifier:
            l = ['identifier="' + self.identifier + '"']
        else:
            l = []
        l += [six.text_type(i) + '="' + six.text_type(j) +
              '"' for i, j in six.iteritems(self.attributes)]
        return '<' + self.type + ' ' + ', '.join(l) + '>'


#------------------------------------------------------------------------------
class EndStructure(object):

    '''
    When iterating over a minf tree, an L{EndStructure} indicate the end of
    a subtree.
    '''

    def __init__(self, structureNodeType):
        self.type = structureNodeType

    def __repr__(self):
        return '</' + self.type + '>'


#------------------------------------------------------------------------------
class Reference(object):

    '''
    EXPERIMENTAl: the reference system is not fully functional.
    When iterating over a minf tree, an L{Reference} correspond to a structure
    previously identified in the minf tree.
    '''

    def __init__(self, identifier):
        '''
        @param identifier: identifier of the referenced object.
        @type identifier: string
        '''
        self.identifier = identifier

    def __repr__(self):
        return '<ref identifier="' + self.identifier + '">'


#------------------------------------------------------------------------------
def createMinfReducer(name):
    '''
    Return an instance of L{MinfReducer} previously registered.

    @param name: name of the reducer
    @type  name: string
    '''
    reducer = MinfReducer._allReducers.get(name)
    if reducer is None:
        raise MinfError(_('Unknown Minf reducer: %s') % (name, ))
    return reducer


#------------------------------------------------------------------------------
class MinfReducer(object):

    '''
    Class to convert a Python object into a minf tree.
    '''
    #: todo: documentation
    _allReducers = {}

    #: todo: documentation
    _defaultClassReducer = {}

    class DefaultObjectReducer(object):

        def __init__(self, structureName):
            self.structureName = structureName

        def __call__(self, reducer, object):
            getinitkwargs = ( getattr(object, '__getnewargs_ex__', None)
                              or getattr(object, '__getinitkwargs__', None))
            if getinitkwargs is None:
                getinitargs = (getattr(object, '__getnewargs__', None)
                               or getattr(object, '__getinitargs__', None))
                if not getinitargs:
                    raise TypeError(
                        'Object of type {0} cannot be serialized by MinfReducer'
                        .format(type(object))
                    )
                args = getinitargs()
                kwargs = {}
            else:
                args, kwargs = getinitkwargs()
            start = StartStructure(self.structureName)
            yield start
            for item in args:
                yield None
                for minfNode in reducer.reduce(item):
                    yield minfNode
            if kwargs:
                for key, value in six.iteritems(kwargs):
                    for minfNode in reducer.reduce(key):
                        yield minfNode
                    for minfNode in reducer.reduce(value):
                        yield minfNode
            end = EndStructure(self.structureName)
            yield end

    def __init__(self, name, bases=()):
        self.name = name
        self.bases = tuple([self._allReducers[i] for i in bases])
        self.typeReducers = {}
        self._allReducers[name] = self

    def getTypeReducer(self, classOrName):
        if not isinstance(classOrName, six.string_types):
            className = classOrName.__module__ + '.' + classOrName.__name__
        else:
            className = classOrName
        reducer = self.typeReducers.get(className)
        if reducer is None:
            for base in self.bases:
                reducer = base.getTypeReducer(className)
                if reducer is not None:
                    break
            else:
                if issubclass(classOrName, HasSignature):
                    return self.hasSignatureReducer
                    raise MinfError(
                        _('Automatic reduction of HasSignature not implemented'))
                raise MinfError(_('Object of type %(class)s cannot be reduced '
                                  'in Minf "%(minf)s" structure') %
                                {'class': className, 'minf': self.name})
        return reducer

    def reduce(self, *args):
        for o in args:
            typeReducer = self.getTypeReducer(o.__class__)
            for minfNode in typeReducer(self, o):
                yield minfNode

    def atomReducer(reducer, atom):
        return (atom, )
    atomReducer = staticmethod(atomReducer)

    def sequenceReducer(reducer, sequence):
        try:
            yield StartStructure(listStructure, length=len(sequence))
        except TypeError:
            yield StartStructure(listStructure)
        for minfNode in reducer.reduce(*sequence):
            yield minfNode
        yield EndStructure(listStructure)
    sequenceReducer = staticmethod(sequenceReducer)

    def dictReducer(reducer, dict):
        try:
            yield StartStructure(dictStructure)
        except TypeError:
            yield StartStructure(dictStructure)
        for key, value in six.iteritems(dict):
            for minfNode in reducer.reduce(key):
                yield minfNode
            for minfNode in reducer.reduce(value):
                yield minfNode
        yield EndStructure(dictStructure)
    dictReducer = staticmethod(dictReducer)

    def hasSignatureNonDefaultValues(o):
        it = six.iteritems(o.signature)
        next(it)
        for key, sigItem in it:
            value = getattr(o, key, Undefined)
            if value is not Undefined \
                    and (value != sigItem.defaultValue
                         or getattr(sigItem, 'writeIfDefault', False)):
                yield (key, value)
    hasSignatureNonDefaultValues = staticmethod(hasSignatureNonDefaultValues)

    def hasSignatureToDict(o):
        d = {}
        for key, value in MinfReducer.hasSignatureNonDefaultValues(o):
            if isinstance(value, HasSignature):
                content = MinfReducer.hasSignatureToDict(value)
                if content:
                    d[key] = content
            else:
                d[key] = value
        return d
    hasSignatureToDict = staticmethod(hasSignatureToDict)

    def hasSignatureReducer(reducer, o):
        for minfNode in reducer.reduce(MinfReducer.hasSignatureToDict(o)):
            yield minfNode
    hasSignatureReducer = staticmethod(hasSignatureReducer)

    def registerAtomType(self, cls):
        className = cls.__module__ + '.' + cls.__name__
        self.typeReducers[
            cls.__module__ + '.' + cls.__name__] = self.atomReducer
        self._defaultClassReducer[className] = self.name

    def registerClass(self, cls, reducer):
        className = cls.__module__ + '.' + cls.__name__
        self.typeReducers[className] = reducer
        self._defaultClassReducer[className] = self.name

    def defaultReducer(value):
        '''
        Return a minf reducer that can reduce C{value}.

        @returns: string or None
        '''
        if isinstance(value, six.class_types):
            # value is a class
            cls = value
        else:
            cls = value.__class__
        className = cls.__module__ + '.' + cls.__name__
        return MinfReducer._defaultClassReducer.get(className, None)
    defaultReducer = staticmethod(defaultReducer)


#------------------------------------------------------------------------------
def createMinfExpander(name):
    '''
    Return an instance of L{MinfExpander} previously registered.

    @param name: name of the expander
    @type  name: string
    '''
    expander = MinfExpander._allExpanders.get(name)
    if expander is None:
        raise KeyError(_('Unknown Minf expander: %s') % (name, ))
    return expander


#------------------------------------------------------------------------------
class MinfExpander(object):

    '''
    Class to convert a minf tree into a Python object.
    '''
    _allExpanders = {}

    class DefaultObjectExpander(object):

        def __init__(self, factory):
            self.factory = factory

        def __call__(self, expander, minfNode, minfNodeIterator, target, targetType, stop_on_error=True, exceptions=[]):
            structureName = minfNode.type
            args = []
            kwargs = {}
            for minfNode in minfNodeIterator:
                if isinstance(minfNode, EndStructure):
                    if minfNode.type != structureName:
                        raise MinfError(_('Wrong Minf structure ending, expecting %(exp)s instead of %(rcv)s') %
                                        {'exp': structureName, 'rcv': minfNode.type})
                    break
                else:
                    key = expander.expand(
                        minfNodeIterator, minfNode, stop_on_error=stop_on_error, exceptions=exceptions)
                    try:
                        value = expander.expand(
                            minfNodeIterator, stop_on_error=stop_on_error, exceptions=exceptions)
                        if key is None:
                            args.append(value)
                        else:
                            kwargs[str(key)] = value
                    except Exception as e:
                        if stop_on_error:
                            raise e
                        else:
                            exceptions.append(sys.exc_info())
            return self.factory(*args, **kwargs)

    def __init__(self, name, bases=()):
        self.name = name
        self.bases = tuple([self._allExpanders[i] for i in bases])
        self.typeExpanders = {}
        self._allExpanders[name] = self
        self.objectsWithIdentifier = {}

    def getTypeExpander(self, structureName):
        expander = self.typeExpanders.get(structureName)
        if expander is None:
            for base in self.bases:
                expander = base.getTypeExpander(structureName)
                if expander is not None:
                    break
            else:
                raise MinfError(_('Minf structure %(struct)s cannot be expanded '
                                  'from Minf "%(minf)s" structure') %
                                {'struct': structureName, 'minf': self.name})
        return expander

    def expand(self, minfNodeIterator, minfNode=Undefined, target=None,
               targetType=Undefined, stop_on_error=True, exceptions=[]):
        if minfNode is Undefined:
            minfNode = next(minfNodeIterator)
        if isinstance(minfNode, StartStructure):
            identifier = minfNode.identifier
            typeExpander = self.getTypeExpander(minfNode.type)
            try:
                result = typeExpander(
                    self, minfNode, minfNodeIterator, target=target,
                    targetType=targetType, stop_on_error=stop_on_error, exceptions=exceptions)
            except Exception as e:
                if stop_on_error:
                    raise
                else:
                    result = None
                    exceptions.append(sys.exc_info())
            if identifier is not None:
                self.objectsWithIdentifier[identifier] = result
            return result
        elif isinstance(minfNode, EndStructure):
            raise MinfError(_('Minf structure %s ended but not started') %
                             (minfNode.type, ))
        elif isinstance(minfNode, Reference):
            return self.objectsWithIdentifier[minfNode.identifier]
        else:
            return minfNode

    def sequenceExpander(expander, minfNode, minfNodeIterator, target,
                         targetType, stop_on_error=True, exceptions=[]):
        if target is None:
            result = []
        else:
            result = target
            while len(result) != 0:
                result.pop()
        if isinstance(targetType, Sequence) and targetType.elementType.mutable:
            length = minfNode.attributes.get('length')
            if length and len(result) < int(length):
                result += [targetType.elementType.createValue()
                           for i in range(int(length) - len(result))]
        itTarget = iter(result)
        for minfNode in minfNodeIterator:
            if isinstance(minfNode, EndStructure):
                if minfNode.type != listStructure:
                    raise MinfError(_('Wrong Minf structure ending, expecting %(exp)s instead of %(rcv)s') %
                                    {'exp': listStructure, 'rcv': minfNode.type})
                break
            else:
                target = None
                if itTarget is not None:
                    try:
                        target = next(itTarget)
                    except StopIteration:
                        itTarget = None
                if target is not None:
                    r = expander.expand(minfNodeIterator, minfNode,
                                        target=target,
                                        targetType=targetType.elementType,
                                        stop_on_error=stop_on_error,
                                        exceptions=exceptions)
                    result.append(r)
                else:
                    try:
                        result.append(
                            expander.expand(minfNodeIterator, minfNode, stop_on_error=stop_on_error, exceptions=exceptions))
                    except Exception as e:
                        if stop_on_error:
                            raise e
                        else:
                            result.append(None)
                            exceptions.append(sys.exc_info())
        return result
    sequenceExpander = staticmethod(sequenceExpander)

    def dictExpander(expander, minfNode, minfNodeIterator, target, targetType, stop_on_error=True, exceptions=[]):
        if target is None:
            result = {}
        else:
            result = target
        for minfNode in minfNodeIterator:
            if isinstance(minfNode, EndStructure):
                if minfNode.type != dictStructure:
                    raise MinfError(_('Wrong Minf structure ending, expectinf %(exp)s instead of %(rcv)s') %
                                    {'exp': dictStructure, 'rcv': minfNode.type})
                break
            else:
                key = expander.expand(
                    minfNodeIterator, minfNode, stop_on_error=stop_on_error, exceptions=exceptions)
                if isinstance(key, list):
                    # list objects are unhashable and cannot be used as dictionary key
                    # in this case they are converted to tuple
                    key = tuple(key)
                if isinstance(result, HasSignature):
                    targetType = result.signature.get(key, Undefined)
                    if targetType is not Undefined:
                        targetType = targetType.type
                    target = getattr(result, key, None)
                    try:
                        if isinstance(target, HasSignature) or isinstance(targetType, Sequence):
                            value = expander.expand(
                                minfNodeIterator, target=target,
                                targetType=targetType, stop_on_error=stop_on_error, exceptions=exceptions)
                        else:
                            value = expander.expand(
                                minfNodeIterator, stop_on_error=stop_on_error, exceptions=exceptions)
                        setattr(result, key, value)
                    except Exception as e:
                        if stop_on_error:
                            raise
                        else:
                            exceptions.append(sys.exc_info())

                else:
                    try:
                        value = expander.expand(
                            minfNodeIterator, stop_on_error=stop_on_error, exceptions=exceptions)
                        result[key] = value
                    except Exception as e:
                        if stop_on_error:
                            raise
                        else:
                            exceptions.append(sys.exc_info())

        return result
    dictExpander = staticmethod(dictExpander)

    def registerStructure(self, typeName, expander):
        self.typeExpanders[typeName] = expander


#------------------------------------------------------------------------------
def createReducerAndExpander(name, *bases):
    '''
    Create a new L{MinfReducer}/L{MinfExpander} pair.

    @param name: name registered and used by L{createMinfReducer} and
      L{createMinfExpander}. By convention this name should match the pattern
      C{<name>_<majorVersion>.<minorVersion>} where C{<name>} is a label
      identifying the content of the minf file, and
      C{<majorVersion>.<minorVersion>} is a version number. When a modification
      is done on the structure of a minf file, the version should increase. If
      the modification is backward compatible C{<minorVersion>} should be
      incremented, otherwise C{<majorVersion>} should be incremented.
      There is a built-in L{MinfReducer}/L{MinfExpander} pair named C{minf_2.0}.
    @type  name: string
    @param bases: L{MinfReducer} (respectively L{MinfExpander}) instances can
      inherit from other instances. C{bases} parameter contains the names of the
      base L{MinfReducer} (respectively L{MinfExpander}).
    @type  bases: tuple containing strings
    '''
    reducer = MinfReducer(name, bases)
    expander = MinfExpander(name, bases)
    return (reducer, expander)


#------------------------------------------------------------------------------
def registerClass(reduction, classToRegister, structureName):
    '''
    Register a class to allow reading and writing instances of this class in a
    minf file.

    @param reduction: name of the L{MinfReducer}/L{MinfExpander} pair in which the
      class is to be registered.
    @type  reduction: string
    @param classToRegister: class to register
    @type  classToRegister: class
    @param structureName: name used to identify the class in the minf file
    @type  structureName: string
    '''
    reducer = createMinfReducer(reduction)
    expander = createMinfExpander(reduction)
    reducer.registerClass(classToRegister,
                          MinfReducer.DefaultObjectReducer(structureName))
    expander.registerStructure(structureName,
                               MinfExpander.DefaultObjectExpander(classToRegister))


#------------------------------------------------------------------------------
def registerClassAs(reduction, subclassToRegister, registeredBaseClass):
    '''
    Register a class as written in a minf file as one of its base class. If you
    register a base class, it does not automatically allow the writing of objects
    derived from this base class to be written in a minf file. To allow this, one
    must either call L{registerClass} to register a new C{structureName} for the
    derived class, or call L{registerClassAs} to saved derived class instances as
    instances of the base class.
    @param reduction: name of the L{MinfReducer}/L{MinfExpander} pair in which
      C{subClassToRegister} is to be registered.
    @type  reduction: string
    @param subclassToRegister: class to register.
    @type  subclassToRegister: class
    @param registeredBaseClass: one of the base class of C{subclassToRegister},
      that must have been already registered in the C{reduction}
      {MinfReducer}/L{MinfExpander} pair. Any instance of
      C{subclassToRegister} will be reduced as an instance of
      C{registeredBaseClass}.
    @type  registeredBaseClass: class
    '''
    reducer = createMinfReducer(reduction)
    reducer.registerClass(subclassToRegister,
                          reducer.getTypeReducer(registeredBaseClass))



# -*- coding: utf-8 -*-
from __future__ import print_function

from __future__ import absolute_import
import unittest
import shutil
import os
import tempfile
import soma.minf.api as minf



class TestMinfIO(unittest.TestCase):
    """ test .minf IO
    """
    def setUp(self):
        self.directory = tempfile.mkdtemp()

    def tearDown(self):
        #print('dir:', self.directory)
        shutil.rmtree(self.directory)

    def test_minf_xml_io(self):
        d = {
            'titi': {'bubu': '50', 'turlute': 12},
            'toto': 'val"u\'e',
            'tutu': [0, 1, 2, [u'papa', 5]]}
        minf_file = os.path.join(self.directory, 'minf_xml_file.minf')
        minf.writeMinf(minf_file, (d,))
        dd = minf.readMinf(minf_file)
        self.assertEqual(d, dd[0])

    def test_minf_py_io(self):
        d = {
            'titi': {'bubu': '50', 'turlute': 12},
            'toto': 'val"u\'e',
            'tutu': [0, 1, 2, [u'papa', 5]]}
        minf_file = os.path.join(self.directory, 'minf_py_file.minf')
        open(minf_file, 'w').write('attributes = ' + repr(d) + '\n')
        dd = minf.readMinf(minf_file)
        self.assertEqual(d, dd[0])


def test():
    suite = unittest.TestLoader().loadTestsFromTestCase(TestMinfIO)
    runtime = unittest.TextTestRunner(verbosity=2).run(suite)
    return runtime.wasSuccessful()


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL license version 2 under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL license version 2 as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license version 2 and that you accept its terms.

"""
This module enables to make a function to be executed in qt thread (main thread).
It is useful when you want to call qt functions from another thread.
It enables to do thread safe calls because all tasks sent are executed in the same thread (qt main thread).

* author: Dominique Geffroy
* organization: NeuroSpin
* license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
"""
from __future__ import print_function

from __future__ import absolute_import
__docformat__ = "restructuredtext en"

import sys
import threading
import six
from soma.qt_gui.qt_backend.QtCore import QObject, QTimer, QEvent, QCoreApplication
from soma import singleton


class FakeQtThreadCall(QObject):

    '''
    Fake QtThreadCall that behave as if always used from main thread.
    '''
    def isInMainThread():
        return True
    isInMainThread = staticmethod(isInMainThread)

    def push(function, *args, **kwargs):
        if kwargs is None or not kwargs:
            function(*args)
        else:
            function(*args, **kwargs)
    push = staticmethod(push)

    def call(function, *args, **kwargs):
        if kwargs is None or not kwargs:
            return function(*args)
        else:
            return function(*args, **kwargs)
    call = staticmethod(call)


class QtThreadCall(singleton.Singleton, QObject):

    """
    This object enables to send tasks to be executed by qt thread (main thread).
    This object must be initialized in qt thread.
    It starts a QTimer and periodically execute tasks in its actions list.

    Attributes
    ----------
    lock: RLock
        lock to prevent concurrent access to actions list
    actions: list
        tasks to execute
    mainThread: Thread
        current thread at object initialisation
    timer: QTimer
        timer to wake this object periodically
    """

    def __singleton_init__(self):
        # QObject.__init__( self, None )
        super(QtThreadCall, self).__singleton_init__(None)
        self.lock = threading.RLock()
        self.actions = []
        # look for the main thread
        mainthreadfound = False
        for thread in threading.enumerate():
            if isinstance(thread, threading._MainThread):
                mainthreadfound = True
                self.mainThread = thread
                break
        if not mainthreadfound:
            print('Warning: main thread not found')
            self.mainThread = threading.current_thread()

    def _postEvent(self):
        class QtThreadCallEvent(QEvent):

            def __init__(self, qthreadcall):
                QEvent.__init__(self, QEvent.Type(QEvent.User + 24))
                self.qthreadcall = qthreadcall
        QCoreApplication.instance().postEvent(self,
                                              QtThreadCallEvent(self))

    def event(self, e):
        sys.stdout.flush()
        if e.type() == QEvent.User + 24:
            self.doAction()
            return True
        return QObject.event(self, e)

    def push(self, function, *args, **kwargs):
        """
        Add a function call to the actions list. the call is executed immediatly if current thread is main thread. Otherwise it will be executed some time in the main thread (asynchronously to the current thread). The function return value is ignored and will be lost.

        Parameters
        ----------
        function: function
            the function to call in main thread.
        """
        if self.isInMainThread():
            if kwargs is None or len(kwargs) == 0:
                function(*args)
            else:
                function(*args, **kwargs)
        else:
            self.lock.acquire()
            try:
                self.actions.append((function, args, kwargs))
                self._postEvent()
            finally:
                self.lock.release()

    def call(self, function, *args, **kwargs):
        """
        Send the function call to be executed in the qt main thread and wait for the result. The result will be returned to the calling thread.

        .. warning::

            If returned objects are thread-depenendent (typically, Qt widgets),
            they must be destroyed within the thread which created them, namely
            the main thread. The :class:`MainThreadLife` object wrapper may be
            helpful for this.

        Parameters
        ----------
        function: function
            the function to call in main thread.

        Returns
        -------
        function call result
        """
        if self.isInMainThread():
            if kwargs is None or len(kwargs) == 0:
                return function(*args)
            return function(*args, **kwargs)
        else:
            semaphore = threading.Semaphore(0)
            self.lock.acquire()
            try:
                self.actions.append(
                    (self._callAndWakeUp, (semaphore, function, args, kwargs), {}))
                self._postEvent()
            finally:
                self.lock.release()
            semaphore.acquire()
                              # block until semaphore is released in
                              # _callAndWakeUp method
            result = semaphore._mainThreadActionResult
            exception = semaphore._mainThreadActionException
            if exception is not None:
                six.reraise(*exception)
            return result

    def _callAndWakeUp(self, semaphore, function, args, kwargs):
        """
        Call the function, set the result in semaphore attributes and release the semaphore.

        Parameters
        ----------
        semaphore: threading.Semaphore
            thread which has added this task is blocked on this semaphore. function call's result will be kept in this semaphore attributes.
        function: function
            the function to call in main thread.
        """
        semaphore._mainThreadActionResult = None
        semaphore._mainThreadActionException = None
        try:
            if kwargs is None or len(kwargs) == 0:
                semaphore._mainThreadActionResult = function(*args)
            else:
                semaphore._mainThreadActionResult = function(*args, **kwargs)
        except:  # noqa: E722
            semaphore._mainThreadActionException = sys.exc_info()
        semaphore.release()
                          # release the semaphore to unblock the thread which
                          # waits for the function call's result

    def isInMainThread(self):
        """
        Returns
        -------
        boolean: True if the current thread is the main thread
        """
        # return threading.currentThread().getName() == 'MainThread'
        return threading.current_thread() is self.mainThread

    def doAction(self):
        """
        This method is called each time the timer timeout.
        It executes all functions in actions list.
        """
        self.lock.acquire()
        try:
            actions = self.actions
            # print("actions to do", self.actions)
            self.actions = []
        finally:
            self.lock.release()
        for (function, args, kwargs) in actions:
            try:
                if kwargs is None or len(kwargs) == 0:
                    function(*args)
                else:
                    function(*args, **kwargs)
            except:  # noqa: E722
                # Should call a customizable function here
                raise


class MainThreadLife(object):
    '''This wrapper class ensures the contained object is deleted in the main
    thread, and not in the current non-GUI thread. The principle is the
    following:

    * acquire a lock
    * pass the object to something in the main thread
    * the main thread waits on the lock while holding a reference on the object
    * we delete the object in the calling thread
    * the lock is releasd from the calling thread
    * now the main thread can go on, and del / release the ref on the object:
      it is the last ref on it, so it is actually deleted there.

    .. warning::

        The thing is only working if no other reference is held anywhere on the underlying object, otherwise we do not control its deletion.

    Ex:

    ::

        # from a secondary thread

        widget = QtThreadCall().call(Qt.QWidget)
        # widget.show() # should crash
        QtThreadCall().push(widget.show)
        # ... use it ...
        # del widget # should crash

    ::

        # from a secondary thread

        widget = MainThreadLife(QtThreadCall().call(Qt.QWidget))
        QtThreadCall().push(widget.ref().show)
        # ... use it ...
        del widget # OK
    '''
    def __init__(self, obj_life=None, *args, **kwargs):
        super(MainThreadLife, self).__init__(*args, **kwargs)
        if obj_life is not None:
            self._obj_life = obj_life

    def __del__(self):
        if (not isinstance(threading.current_thread(), threading._MainThread)
                and hasattr(self, '_obj_life')):
            lock = threading.Lock()
            lock.acquire()
            QtThreadCall().push(MainThreadLife.delInMainThread, lock,
                                self._obj_life)
            del self._obj_life
            lock.release()

    def ref(self):
        '''Access the underlying object'''
        return self._obj_life

    @staticmethod
    def delInMainThread(lock, thing):
        # wait for the lock to be released in the process thread
        lock.acquire()
        lock.release()
        # now the process thread should have removed its reference on thing:
        # we can safely delete it fom here, in the main thread.
        del thing # probably useless



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
Utility classes and functions for Python import and sip namespace renaming.

* author: Yann Cointepas
* organization: NeuroSpin
* license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
'''
from __future__ import absolute_import
__docformat__ = "restructuredtext en"

import sys
import imp
import types
import six
import importlib

from soma.functiontools import partial
from soma.singleton import Singleton


# list of namespace objects that should not be patched to avoid a side effect
# in sip imported namespaces: we must not access their attributes during
# imports.
__namespaces__ = ['soma', 'aims', 'carto', 'anatomist']


class ExtendedImporter(Singleton):

    '''
      ExtendedImporter is used to import external modules in a module managing rules that allow ro rename and delete or do anything else on the imported package. As imported packages could modify each others, all the registered rules are applied after each import using this ExtendedImporter.
    '''
    extendedModules = dict()

    def importInModule(self, moduleName, globals, locals, importedModuleName,
                       namespacesList=[], handlersList=None, *args, **kwargs):
        '''
        This method is used to import a module applying rules (rename rule, delete rule, ...) .

        Parameters
        ----------
        moduleName: string
            name of the module to import into (destination, not where to find
            it).
        globals: dict
            globals dictionary of the module to import into.
        locals: dict
            locals dictionary of the module to import into.
        importedModuleName: string
            name of the imported module. Normally relative to the current
            calling module package.
        namespacesList: list
            a list of rules concerned namespaces for the imported module.
        handlersList: list
            a list of handlers to apply during the import of the module.
        '''

        if (handlersList is None):
            # Add default handler
            handlersList = [GenericHandlers.moveChildren]

        if not moduleName:
            moduleName = locals['__name__']
        elif moduleName.startswith('.'):
            moduleName = locals['__name__'] + moduleName

        package = locals['__package__'] or locals['__name__']

        # Import the module
        # Note : Pyro overloads __import__ method and usual keyword 'level' of
        # __builtin__.__import__ is not supported
        importedModule = importlib.import_module('.' + importedModuleName,
                                                 package)
        sys.modules[importedModuleName.split('.')[-1]] = importedModule

        # Add the extended module to the list if not already exists
        if moduleName not in self.extendedModules:
            extendedModule = ExtendedModule(moduleName, globals, locals)
            self.extendedModules[moduleName] = extendedModule
        else:
            extendedModule = self.extendedModules[moduleName]

        if len(namespacesList) == 0:
            extendedModule.addHandlerRules(
                importedModule, handlersList, importedModuleName, *args,
                **kwargs)
        else:
            for namespace in namespacesList:
                extendedModule.addHandlerRules(
                    importedModule, handlersList, namespace, *args, **kwargs)

        self.applyRules()

    def applyRules(self):
        '''
        This method apply rules for each extended module.
        '''
        for extendedModule in self.extendedModules.values():
            extendedModule.applyRules()


class ExtendedModule(object):

    '''
    Register a series of rules to apply during the import process of the
    extended module. An extended module is able to refer to other modules and
    to apply rules to these other modules. The extended module manages the
    globals and locals variables declared for the module.

    Each rule contains the module to which it refers, and the handlers to call
    for the rule to apply. The calling order is the registering order.
    '''

    def __init__(self, moduleName, globals, locals):
        self.rules = dict()
        self.__name__ = moduleName
        self.globals = globals
        self.locals = locals

    def addHandlerRules(self, module, handlersList=[], *args, **kwargs):
        '''
        This method is used to add handler rules (renaming rule, deleting rule, ...) .

        Parameters
        ----------
        module: module
            module object to apply rules to.
        handlersList: list
            a list of handlers to the module.
        '''
        for handler in handlersList:
            self.addPartialRules(
                module, [partial(handler, *args, **kwargs)])

    def addPartialRules(self, module, partialsList=[]):
        '''
        This method is used to add handler rules (renaming rule, deleting rule, ...) .

        Parameters
        ----------
        module: module
            module object to apply rules to.
        partialsList: list
            a list of :func:`functools.partial` objects that will be called
            during the import of the module.
        '''
        key = module

        if key not in self.rules:
            self.rules[key] = list()

        for handler in partialsList:
            if handler not in self.rules[key]:
                self.rules[key].append(handler)

    def applyRules(self):
        '''
        Apply the :func:`functools.partial` handler rules (renaming rule, deleting rule, ...) for the :class:`ExtendedModule`.
        '''
        for referedModule, partialsList in self.rules.items():
            for handler in partialsList:
                # Call the handlers list for the found object
                handler.__call__(self, referedModule)


class GenericHandlers(object):

    '''
    Static generic handlers used as import rules.
    '''
    @staticmethod
    def moveChildren(namespace, extendedModule, referedModule):
        '''
        This static method is used to move child objects of a refered module to the extended module.

        Parameters
        ----------
        namespace: string
            namespace of the referedModule to get objects to move from.
        extendedModule: ExtendedModule
            :class:`ExtendedModule` object into which move objects.
        referedModule: module
            refered module object to get objects to move from.
        '''

        # during this whole function we should avoid tu use
        # getattr(mobject, something) (directly or indirectly) because it
        # breaks things in sip imports later.

        # Get module names
        newName = extendedModule.__name__

        # Get extended module locals declaration
        locals = extendedModule.locals

        # Get the old object in module
        mobject = ExtendedImporterHelper.getModuleObject(
            referedModule, namespace)

        if mobject is not None:

            # Changes child objects locals declaration
            for childName \
                    in list(object.__getattribute__(mobject, '__dict__').keys()):
                # in sip >= 4.8, obj.__dict__[key] and getattr(obj, key)
                # do *not* return the same thing for functions !
                # object.__getattribute__(mobject, childName) returns a
                # "methoddescriptor", whereas
                # getattr(mobject, childName) returns a
                # "builtin_function_or_method", which is what we want

                # childObject = object.__getattribute__(mobject, childName)
                childObject = getattr(mobject, childName)
                if not childName.startswith("__"):

                    locals[childName] = childObject

            # Changes child objects module, recursively and avoiding loops
            stack = []
            done = []
            for (childName, childObject) in six.iteritems(locals):
                if not childName.startswith("__"):
                    try:
                        mod = object.__getattribute__(childObject,
                                                      '__module__')
                        if mod == referedModule.__name__:
                            stack.append(childObject)
                    except AttributeError:
                        pass

            # set new module on objects

            while stack:
                childObject = stack.pop(0)
                done.append(childObject)
                try:
                    name = object.__getattribute__(childObject, '__name__')
                    mod = object.__getattribute__(childObject, '__module__')
                except AttributeError:
                    continue
                try:
                    # skip namespace objects
                    if name not in __namespaces__:
                        childObject.__module__ = newName
                except Exception as e:
                    pass
                try:
                    d = object.__getattribute__(childObject, '__dict__')
                except AttributeError:
                    continue
                for x in list(d.keys()):
                    try:
                        # y = object.__getattribute__(childObject, x)
                        y = getattr(childObject, x)
                        if not x.startswith( '__' ) \
                                and y not in stack and y not in done:
                            stack.append(y)
                    except AttributeError:
                        pass

    # Declare a function to delete non generics Reader/Writer objects
    @staticmethod
    def removeChildren(namespace, prefixes, extendedModule, referedModule):
        '''
        This static method is used to remove children from the extended module.

        Parameters
        ----------
        namespace: string
            namespace of the referedModule.
        prefixes: list
            list of prefixes of objects to remove.
        extendedModule: ExtendedModule
            :class:`ExtendedModule` object to remove objects from.
        referedModule: module
            refered module object.
        '''
        locals = extendedModule.locals

        # Remove Reader and Writer classes because a generic class
        # to manage it exists
        to_remove = []
        for key in locals.keys():
            if not key.startswith('__'):
                for prefix in prefixes:
                    if key.startswith(prefix):
                        to_remove.append(key)
        for key in to_remove:
            del locals[key]


class ExtendedImporterHelper(object):

    '''
    Static methods declared to help extended import process.
    '''

    @staticmethod
    def getModuleObject(module, name):
        '''
        This static method is used to get an object contained in the namespace of a module.

        Parameters
        ----------
        name: string
            complete name including namespace of the object to get.
        module: module
            module object to get objects from.

        Returns
        -------
        object:
            Object found in the module or None if no object was found.
        '''

        # Split the name of the object
        # but keep the 1st element unsplit like the module name
        # if needed.
        if name == '':
            return module

        names = name.split('.')
        if not hasattr(module, names[0]) and module.__name__.split('.')[-1] == names[0]:
            names = names[1:]
        obj = module
        while names:
            mname = names.pop(0)
            if not hasattr(obj, mname):
                return None
            obj = getattr(obj, mname)
        return obj


def execfile(filename, globals=None, locals=None):
    ''' Replacement for python2 execfile()
    six.exec_() needs an open file, hence this wrapper for convenience.
    Files are open with UTF-8 encoding on python3.
    '''
    fopts = {} if six.PY2 else {'encoding': 'utf-8'}
    with open(filename, **fopts) as f:
        six.exec_(f, globals, locals)



# -*- coding: utf-8 -*-

def dict_to_table(data_dict, shape_dict):
    ''' Tranform a dictionary into a table shape
    (list of list, each sublist having the same size)

    The first line will be the header (name of columns).

    shape_dict will help setting names to columns.

    For instance::

        data_dict = {
            0: {'areas': {0: 23483.013671875,
                          2: 325.46331787109375,
                          7: 1300.6187744140625},
                'lengths': {0: 2225.728,
                            2: 80.755554,
                            7: 323.5833,
                            8: 103.076324}},
            1: {'areas': {2: 350.5135}}
        }
        shape_dict = {'timestep': {'areas': {'parcel': 'area'},
                                   'lengths': {'parcel': 'length'}}}
        dict_to_table(data_dict, shape_dict)

    will output::

        [['timestep', 'parcel', 'area', 'length'],
         [0, 0, 23483.013671875, 2225.728],
         [0, 2, 325.46331787109375, 80.755554],
         [0, 7, 1300.6187744140625, 323.5833],
         [0, 8, None, 103.076324],
         [1, 2, 350.5135, None]]

    At each level of the data dict, some additional, fixed, attributes, may
    be specified in an optional "fixed_attributes" sub-dict. Ex::

        data_dict = {
            0: {'areas': {0: 23483.013671875,
                          2: 325.46331787109375,
                          7: 1300.6187744140625},
                'lengths': {0: 2225.728,
                            2: 80.755554,
                            7: 323.5833,
                            8: 103.076324}},
            1: {'areas': {2: 350.5135}},
            'fixed_attributes': {'subject': 'maman'},
        }

    will output (with the same shape_dict)::

        [['subject', 'timestep', 'parcel', 'area', 'length'],
         ['maman', 0, 0, 23483.013671875, 2225.728],
         ['maman', 0, 2, 325.46331787109375, 80.755554],
         ['maman', 0, 7, 1300.6187744140625, 323.5833],
         ['maman', 0, 8, None, 103.076324],
         ['maman', 1, 2, 350.5135, None]]
    '''

    todo = [(data_dict, shape_dict, {})]
    line_items = []
    while todo:
        data, shape, keys = todo.pop(0)
        if isinstance(data, dict):
            if 'fixed_attributes' in data:
                keys.update(data['fixed_attributes'])
                data = dict(data)
                del data['fixed_attributes']
            if len(shape) == 1:
                key, sshape = next(iter(shape.items()))
                for k, v in data.items():
                    keys = dict(keys)
                    keys[key] = k
                    todo.append((v, sshape, keys))
            else:
                for key, sshape in shape.items():
                    if key in data:
                        todo.append((data[key], sshape, keys))
        else:
            line = dict(keys)
            line[shape] = data  # shape is normally just a string now
            line_items.append(line)

    if len(line_items) == 0:
        return []

    hdr = list(line_items[0].keys())
    keys = {k: i for i, k in enumerate(hdr)}
    for line in line_items:
        for k, v in line.items():
            if k not in hdr:
                keys[k] = len(hdr)
                hdr.append(k)
    lines = []
    for line in line_items:
        l0 = [line.get(k) for k in hdr]
        found = False
        for l in lines:
            eq = True
            for e, v in zip(l, l0):
                if e is not None and v is not None and e != v:
                    eq = False
                    break
            if eq:  # all values present in both are the same
                # merge lines
                found = True
                for i, v in enumerate(l0):
                    if v is not None:
                        l[i] = v
                break
        if not found:
            lines.append(l0)

    return [hdr] + lines



# -*- coding: utf-8 -*-
#
# SOMA - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
#

# System import
from __future__ import print_function
from __future__ import absolute_import
from .List_offscreen import OffscreenListControlWidget
from soma.qt_gui.qt_backend import Qt
from soma.qt_gui import qt_backend
from functools import partial
import os


class OffscreenListFileControlWidget(OffscreenListControlWidget):

    @staticmethod
    def create_widget(parent, control_name, control_value, trait,
                      label_class=None, user_data=None):
        """ Method to create the list widget.

        Parameters
        ----------
        parent: QWidget (mandatory)
            the parent widget
        control_name: str (mandatory)
            the name of the control we want to create
        control_value: list of items (mandatory)
            the default control value
        trait: Tait (mandatory)
            the trait associated to the control
        label_class: Qt widget class (optional, default: None)
            the label widget will be an instance of this class. Its constructor
            will be called using 2 arguments: the label string and the parent
            widget.

        Returns
        -------
        out: 2-uplet
            a two element tuple of the form (control widget: ,
            associated labels: (a label QLabel, the tools QWidget))
        """
        widget, labels = OffscreenListControlWidget.create_widget(
            parent, control_name, control_value, trait,
            label_class=label_class, user_data=user_data)

        button = Qt.QPushButton("...", widget)
        button.setObjectName('files_button')
        button.setStyleSheet('QPushButton#files_button '
                             '{padding: 2px 10px 2px 10px; margin: 0px;}')
        layout = widget.layout()
        layout.addWidget(button)
        button.clicked.connect(partial(
            OffscreenListFileControlWidget.select_files, widget))

        return (widget, labels)


    @staticmethod
    def select_files(control_instance):
        """ Browse the file system and update the control instance accordingly.

        If a valid file path has already been entered the file dialogue will
        automatically point to the file folder, otherwise the current working
        directory is used.

        Parameters
        ----------
        control_instance: QWidget (mandatory)
            the file widget item
        """
        # Get the current file path
        init_dir = os.getcwd()

        # get widget via a __self__ in a method, because control_instance may
        # be a weakproxy.
        widget = control_instance.__repr__.__self__
        ext = []
        trait = control_instance.trait.handler.inner_traits()[0]
        if trait.allowed_extensions:
            ext = trait.allowed_extensions
        if trait.extensions:
            ext = trait.extensions
        ext = ['*%s' % e for e in ext]
        ext = ' '.join(ext)
        if ext:
            ext += ';; All files (*)'
        # Create a dialog to select a file
        if control_instance.trait.output:
            fname = qt_backend.getSaveFileName(
                widget, "Output files", init_dir, ext,
                None, Qt.QFileDialog.DontUseNativeDialog)
            fnames = [fname]
        else:
            fnames = Qt.QFileDialog.getOpenFileNames(
                widget, "Open files", init_dir, ext, None,
                Qt.QFileDialog.DontUseNativeDialog)
            if fnames:
                fnames = fnames[0]

        if fnames:
            # Set the selected files path to the control
            controller = control_instance.parent().controller
            print('controller:', controller)
            print('param:', control_instance.trait_name)
            print('value:', fnames)
            setattr(controller, control_instance.trait_name, fnames)



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
This module contains all the framework necessary to customize, read and
write minf files. A minf file is composed of structured data saved in
XML or Python format. The minf framework provides tools to read and write
minf files but also to customize the way Python objects are read an written.

There are several submodules in this package but main functions and classes
can be imported from :py:mod:`soma.minf.api`:

- for reading minf files: :func:`iterateMinf`, :func:`readMinf`
- for writing minf files: :func:`createMinfWriter`, :func:`writeMinf`
- for customizing minf files: :func:`createReducerAndExpander`, :func:`registerClass`, :func:`registerClassAs`

* author: Yann Cointepas
* organization: NeuroSpin
* license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
'''
from __future__ import print_function
from __future__ import absolute_import
__docformat__ = "restructuredtext en"

import gzip
import six
import sys

from soma.translation import translate as _
from soma.minf.error import MinfError
from soma.bufferandfile import BufferAndFile
from soma.minf.reader import MinfReader
from soma.minf.writer import MinfWriter
from soma.minf.tree import createReducerAndExpander, registerClass, \
    registerClassAs, createMinfExpander, \
    EndStructure, MinfReducer, MinfExpander, \
    listStructure, dictStructure
from soma.undefined import Undefined
defaultReducer = MinfReducer.defaultReducer


#------------------------------------------------------------------------------
def minfFormat(source):
    '''
    Return a pair (format, reduction) identifying the minf format. If
    source is not a minf file, (None, None) is returned. Otherwise,
    format is a string representing the format of the minf file: 'XML' or
    'python'. reduction is the name of the reducer used to write the minf
    file or None if format is 'python'.

    Example:

    ::

      from soma.minf.api import minfFormat
      format, reduction = minfFormat('/home/me/test.minf')

    If source is a :class:`BufferAndFile` instance, this call behave as if nothing
    has been read from the file. This can be useful if you have an opened file
    that cannot be seeked backward:

    Example:

    ::

      from soma.bufferandfile import BufferAndFile
      from soma.minf.api import minfFormat, readMinf

      bf = BufferAndFile(stream_file_object)
      format, reduction = minfFormat(bf)
      if format is not None:
        minfContent = readMinf(bf)


    Parameters
    ----------
    source: string
      Input file name or file object. If it is a file name, it is
      opened with open(source).
    '''
    opened_source_file = None
    try:
        if not hasattr(source, 'readline'):
            if six.PY2:
                opened_source_file = open(source)
            else:
                opened_source_file = open(source, encoding='UTF-8')
                
            source = BufferAndFile(opened_source_file)
        elif not isinstance(source, BufferAndFile):
            source.seek(0)
            source = BufferAndFile(source)

        # Check first non white character to see if the minf file is XML or not
        start = source.read(5)
        if start == 'attri':
            source.unread(start)
            return ('python', None)
        elif start != '<?xml':
            # Try gzip compressed file
            gzipSource = source.clone()
            gzipSource.unread(start)
            gunzipSource = gzip.GzipFile(source.name)
            try:
                start = gunzipSource.read(5)
            except IOError:
                start = ''
            if start != '<?xml':
                raise MinfError(_('Invalid minf file: %s') % (source.name, ))
            source.change_file(gunzipSource)
            source.unread(start)
        else:
            source.unread(start)

        r = MinfReader.createReader('XML')
        reduction, buffer = r.reduction(source)
        source.unread(buffer)
        return('XML', reduction)
    finally:
        if opened_source_file is not None:
            opened_source_file.close()

#------------------------------------------------------------------------------
def _setTarget(target, source):
    try:
        from soma.signature.api import HasSignature
    except ImportError:
        class HasSignature(object):
            pass
    if isinstance(source, dict):
        if isinstance(target, dict):
            target.update(source)
            return True
        elif isinstance(target, HasSignature):
            for k, v in six.iteritems(source):
                attrTarget = getattr(target, k, Undefined)
                if attrTarget is Undefined:
                    setattr(target, k, v)
                else:
                    if not _setTarget(attrTarget, v):
                        setattr(target, k, v)
            return True
    return False


#------------------------------------------------------------------------------
def iterateMinf(source, targets=None, stop_on_error=True, exceptions=[]):
    '''
    Returns an iterator over all objects stored in a minf file.

    Example:

    ::

      from soma.minf.api import iterateMinf

      for item in iterateMinf('test.minf'):
        print(repr(item))

    Parameters
    ----------
    source: string
      Input file name or file object. If it is a file name, it is
      opened with C{open( source )}.
    '''
    if targets is not None:
        targets = iter(targets)

    initial_source = source

    if not hasattr(initial_source, 'readline') and not six.PY2:
        # in python3 the encoding of a file should be specified when opening
        # it: it cannot be changed afterwards. So in python3 we cannot read
        # the encoding within the file (for instance in a XML file).
        # This is completely silly, but here it is...
        # So we just have to try several encodings...
        try_encodings = ['UTF-8', 'latin1']
    else:
        try_encodings = [None]

    for encoding in try_encodings:
        opened_source_file = None
        if not hasattr(initial_source, 'readline'):
            if six.PY2:
                opened_source_file = open(initial_source)
            else:
                opened_source_file = open(initial_source, encoding=encoding)
            source = BufferAndFile(opened_source_file)
        elif not isinstance(source, BufferAndFile):
            source.seek(0)
            source = BufferAndFile(source)

        try:
            # Check first non white character to see if the minf file is XML or not
            start = source.read(5)
            source.unread(start)

            if start == 'attri':
                try:
                    import numpy
                    d = {'nan': numpy.nan}
                except Exception:
                    d = {'nan': None}
                try:
                    six.exec_(source.read().replace("\r\n", "\n"), d)
                except Exception as e:
                    x = source
                    if hasattr(source, '_BufferAndFile__file'):
                        x = source._BufferAndFile__file
                    x = 'Error in iterateMinf while reading ' + str(x) + ': '
                    # e.args = ( x + e.args[0], ) + e.args[1:]
                    print(x)
                    raise
                minf = d['attributes']
                if targets is not None:
                    result = next(targets)
                    _setTarget(result, minf)
                    yield result
                else:
                    yield minf
                return
            elif start != '<?xml':
                # Try gzip compressed file
                gzSource = gzip.GzipFile(source.name)
                if gzSource.read(5) != '<?xml':
                    raise MinfError(_('Invalid minf file: %s')
                                    % (source.name, ))
                source = BufferAndFile(gzSource)
                source.unread('<?xml')

            r = MinfReader.createReader('XML')
            iterator = r.nodeIterator(source)
            minfNode = next(iterator)
            expander = createMinfExpander(minfNode.attributes['reduction'])
            count = 0
            for nodeItem in iterator:
                count += 1
                if isinstance(nodeItem, EndStructure):
                    break
                target = None
                if targets is not None:
                    try:
                        target = next(targets)
                    except StopIteration:
                        targets = None
                yield expander.expand(iterator, nodeItem, target=target,
                                      stop_on_error=stop_on_error,
                                      exceptions=exceptions)
        except UnicodeDecodeError as e:
            if encoding == try_encodings[-1]:
                raise
            continue
        finally:
            if opened_source_file is not None:
                opened_source_file.close()
        break # no error, don't process next encoding

#------------------------------------------------------------------------------


def readMinf(source, targets=None, stop_on_error=True, exceptions=[]):
    '''
    Entirerly reads a minf file and returns its content in a tuple.
    Equivalent to tuple(iterateMinf(source)).

    see: :func`iterateMinf`
    '''
    return tuple(iterateMinf(source, targets=targets,
                             stop_on_error=stop_on_error,
                             exceptions=exceptions))


#------------------------------------------------------------------------------
def createMinfWriter(destFile, format='XML', reducer='minf_2.0'):
    '''
    Create a writer for storing objects in destFile.
    Example:

    ::

      from soma.minf.api import createMinfWriter
      writer = createMinfWriter( '/tmp/test.minf' )
      writer.write( 'A string' )
      writer.write( { 'A dict key': [ 'A list', 'with two elements' ] } )
      writer.close()

    Parameters
    ----------
    format: string
      name of the format to write.
    reducer: string
      name of the reducer to use (see L{soma.minf.tree} for
      more information about reducers).
    '''
    return MinfWriter.createWriter(destFile, format, reducer)


#------------------------------------------------------------------------------
def writeMinf(destFile, args, format='XML', reducer=None):
    '''
    Creates a minf writer with :func:`createMinfWriter` and write the content
    of args in it.

    Parameters
    ----------
    destFile:
      see :func:`createMinfWriter`
    args: sequence or iterator
      series of values to write
    format:
      see :func:`createMinfWriter`
    reducer:
      see :func:`createMinfWriter`
    '''
    it = iter(args)
    try:
        firstItem = next(it)
    except StopIteration:
        firstItem = Undefined
    if reducer is None:
        if firstItem is not Undefined:
            reducer = MinfReducer.defaultReducer(firstItem)
            if reducer is None:
                reducer = 'minf_2.0'

    writer = createMinfWriter(destFile, format, reducer)
    if firstItem is not Undefined:
        writer.write(firstItem)
        for item in it:
            writer.write(item)
    writer.close()


#------------------------------------------------------------------------------
# xml_reader and xml_writer are not used directly but importing
# them register the XML minf format
import soma.minf.xml_reader
import soma.minf.xml_writer


#------------------------------------------------------------------------------
from soma.minf.xhtml import XHTML
from soma.uuid import Uuid

minf_2_0_reducer = MinfReducer('minf_2.0')
minf_2_0_reducer.registerAtomType(None.__class__)
minf_2_0_reducer.registerAtomType(bool)
for t in six.integer_types:
    minf_2_0_reducer.registerAtomType(t)
minf_2_0_reducer.registerAtomType(float)
minf_2_0_reducer.registerAtomType(str)
minf_2_0_reducer.registerAtomType(six.text_type)  # unicode in Python 2
minf_2_0_reducer.registerAtomType(XHTML)
minf_2_0_reducer.registerClass(list, minf_2_0_reducer.sequenceReducer)
minf_2_0_reducer.registerClass(tuple, minf_2_0_reducer.sequenceReducer)
minf_2_0_reducer.registerClass(dict, minf_2_0_reducer.dictReducer)

minf_2_0_expander = MinfExpander('minf_2.0')
minf_2_0_expander.registerStructure(
    listStructure, minf_2_0_expander.sequenceExpander)
minf_2_0_expander.registerStructure(
    dictStructure, minf_2_0_expander.dictExpander)

registerClass('minf_2.0', Uuid, 'Uuid')


#------------------------------------------------------------------------------
minf_1_0_reducer = MinfReducer('minf_1.0', ('minf_2.0', ))
minf_1_0_expander = MinfExpander('minf_1.0', ('minf_2.0', ))



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
Utility classes and functions for Python callable.

* author: Yann Cointepas
* organization: `NeuroSpin <http://www.neurospin.org>`_
* license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
'''
from __future__ import absolute_import
from six.moves import zip
import six
__docformat__ = "restructuredtext en"

import inspect
# handle deprecation / removal of getargspec in python3
getfullargspec = getattr(inspect, 'getfullargspec',
                         getattr(inspect, 'getargspec', None))

#-------------------------------------------------------------------------
from soma.translation import translate as _

#-------------------------------------------------------------------------
class Empty(object):
    pass

#-------------------------------------------------------------------------


class SomaPartial(object):

    '''
    This is a reimplementation of :py:func:`functools.partial`, which adds
    compatibility with the :py:mod:`traits` module.

    See the documentation of :py:func:`functools.partial` for details:
    https://docs.python.org/library/functools.html#functools.partial

    Example::

        from soma.functiontools import partial

        def f(a, b, c):
            return (a, b, c)

        g = partial(f, 'a', c='c')
        g('b') # calls f('a', 'b', c='c')
    '''

    def __init__(self, function, *args, **kwargs):
        self.func = function
        self.args = args
        self.keywords = kwargs

    def __call__(self, *args, **kwargs):
        merged_kwargs = self.keywords.copy()
        merged_kwargs.update(kwargs)
        return self.func(*(self.args + args), **merged_kwargs)

    @property
    def func_code(self):
        '''
        This property make SomaPartial useable with traits module. The method
        on_trait_change is requiring that the function has a func_code.co_argcount
        attribute.
        '''
        result = Empty()
        result.co_argcount = numberOfParameterRange(self)[0]
        return result

    @property
    def __code__(self):
        return self.func_code

try:
    from functools import partial
except ImportError:
    partial = SomaPartial

#-------------------------------------------------------------------------


def getArgumentsSpecification(callable):
    '''
    This is an extension of Python module :py:mod:`inspect.getfullargspec` that
    accepts classes and returns only information about the parameters that can
    be used in a call to *callable* (*e.g.* the first *self* parameter of bound
    methods is ignored). If *callable* has not an appropriate type, a
    :class:`TypeError <exceptions.TypeError>` exception is raised.

    Parameters
    ----------
    callable: callable
        *function*, *method*, *class* or *instance* to inspect

    Returns
    -------
    tuple:
        As :func:`inspect.getfullargspec`, returns
        *(args, varargs, varkw, defaults)* where *args* is a list of the
        argument names (it may contain nested lists). *varargs* and *varkw* are
        the names of the ``*`` and ``**`` arguments or *None*. *defaults* is a
        n-tuple of the default values of the last *n* arguments.
    '''
    if inspect.isfunction(callable):
        return getfullargspec(callable)[:4]
    elif inspect.ismethod(callable):
        args, varargs, varkw, defaults = getfullargspec(callable)[:4]
        args = args[1:]  # ignore the first "self" parameter
        return args, varargs, varkw, defaults
    elif inspect.isclass(callable):
        try:
            init = callable.__init__
        except AttributeError:
            return [], None, None, None
        return getArgumentsSpecification(init)
    elif isinstance(callable, (partial, SomaPartial)):
        args, varargs, varkw, defaults = getArgumentsSpecification(
            callable.func)
        if defaults:
            d = dict(zip(reversed(args), reversed(defaults)))
        else:
            d = {}
        d.update(zip(reversed(args), reversed(callable.args)))
        if callable.keywords:
            d.update(callable.keywords)

        if len(d):
            defaults = tuple((d[i] for i in args[-len(d):]))
        else:
            defaults = d

        return (args, varargs, varkw, defaults)
    else:
        try:
            call = callable.__call__
        except AttributeError:
            raise TypeError(_('%s is not callable') %
                            repr(callable))
        return getArgumentsSpecification(call)

#-------------------------------------------------------------------------


def getCallableString(callable):
    '''
    Returns a translated human readable string representing a callable.

    Parameters
    ----------
    callable: callable
        *function*, *method*, *class* or *instance* to inspect

    Returns
    --------
    string:
        type and name of the callable
    '''
    if inspect.isfunction(callable):
        name = _('function %s') % (callable.__name__, )
    elif inspect.ismethod(callable):
        name = _('method %s') % (six.get_method_self(callable).__class__.__name__ + '.' +
                                 callable.__name__, )
    elif inspect.isclass(callable):
        name = _('class %s') % (callable.__name__, )
    else:
        name = str(callable)
    return name

#-------------------------------------------------------------------------


def hasParameter(callable, parameterName):
    '''
    Returns *True* if *callable* can be called with a parameter named
    *parameterName*. Otherwise, returns *False*.

    .. seealso:: :py:func:`getArgumentsSpecification`

    Parameters
    ----------
    callable: callable
        *function*, *method*, *class* or *instance* to inspect
    parameterName: string
        name of the parameter

    Returns
    -------
    bool:

    '''
    args, varargs, varkw, defaults = getArgumentsSpecification(callable)
    return varkw is not None or parameterName in args

#-------------------------------------------------------------------------


def numberOfParameterRange(callable):
    '''
    Returns the minimum and maximum number of parameter that can be used to
    call a function. If the maximum number of argument is not defined, it is
    set to *None*.

    .. seealso:: :func:`getArgumentsSpecification`

    Parameters
    ----------
    callable: callable
        *function*, *method*, *class* or *instance* to inspect

    Returns
    -------
    tuple:
        two elements (minimum, maximum)

    '''
    args, varargs, varkw, defaults = getArgumentsSpecification(callable)
    if defaults is None or len(defaults) > len(args):
        lenDefault = 0
    else:
        lenDefault = len(defaults)
    minimum = len(args) - lenDefault
    if varargs is None:
        maximum = len(args)
    else:
        maximum = None
    return minimum, maximum


#-------------------------------------------------------------------------
def checkParameterCount(callable, paramCount):
    '''
    Checks that a callable can be called with *paramCount* arguments. If not, a
    RuntimeError is raised.

    .. seealso:: :func:`getArgumentsSpecification`

    Parameters
    ----------
    callable: callable
        *function*, *method*, *class* or *instance* to inspect
    paramCount: int
        number of parameters

    '''
    minimum, maximum = numberOfParameterRange(callable)
    if ( maximum is not None and paramCount > maximum ) or \
       paramCount < minimum:
        raise RuntimeError(
            _('%(callable)s cannot be called with %(paramCount)d arguments') %
            {'callable': getCallableString(callable),
             'paramCount': paramCount})

#-------------------------------------------------------------------------


def drange(start, stop, step=1):
    '''
    Creates lists containing arithmetic progressions of any number type (int,
    float, ...)
    '''
    r = start
    while r < stop:
        yield r
        r += step



# -*- coding: utf-8 -*-

import inspect
import enum
import sys


def sip4_to_sip6_enums(module, recursive=True):
    ''' Convert Sip4 style enums to sip6 style enums.

    Sip4 exports C++ enums as in C++, ex::

        instance.ENUM_VALUE  # where ENUM_VALUE is an instance of EnumType

    Sip6 keeps values inside the enum type::

        instance.EnumType.ENUM_VALUE

    In order to maintain code compatible, we duplicat these values. If using
    sip4, then we copy values inside the enum types as sip6 does.

    If recursive (the default), sub-modules are also scanned and modified.
    '''
    todo = [module]
    done = set()
    while todo:
        module = todo.pop(0)
        if id(module) in done:
            continue

        done.add(id(module))
        enums = set()
        for iname, item in module.__dict__.items():
            if type(item).__name__ == 'enumtype' and id(item) not in done:
                enums.add(item)
                done.add(id(item))
            elif inspect.isclass(item) or (recursive
                                           and inspect.ismodule(item)):
                todo.append(item)
        if enums:
            for iname, item in module.__dict__.items():
                if type(item) in enums:
                    setattr(type(item), iname, item)


def sip6_to_sip4(module, recursive=True):
    ''' Convert Sip6 style enums to sip4 style enums.

    Sip4 exports C++ enums as in C++, ex::

        instance.ENUM_VALUE  # where ENUM_VALUE is an instance of EnumType

    Sip6 keeps values inside the enum type::

        instance.EnumType.ENUM_VALUE

    In order to maintain code compatible, we duplicat these values. If using
    sip6, then we copy values outside the enum types as sip4 does.

    If recursive (the default), sub-modules are also scanned and modified.
    '''
    todo = [module]
    done = set()
    while todo:
        module = todo.pop(0)
        if id(module) in done:
            continue

        done.add(id(module))
        enums = []
        for iname, item in module.__dict__.items():
            if type(item) is enum.EnumMeta and id(item) not in done:
                enums.append(item)
                done.add(id(item))
            elif inspect.isclass(item) or (recursive
                                           and inspect.ismodule(item)):
                todo.append(item)
        for entype in enums:
            for name in entype._member_names_:
                setattr(module, name, getattr(entype, name))


def sip_export_enums(module, recursive=True):
    ''' Convert Sip6 style enums to sip4 style enums, or the contrary.

    Sip4 exports C++ enums as in C++, ex::

        instance.ENUM_VALUE  # where ENUM_VALUE is an instance of EnumType

    Sip6 keeps values inside the enum type::

        instance.EnumType.ENUM_VALUE

    In order to maintain code compatible, we duplicat these values. If using
    sip6, then we copy values outside the enum types as sip4 does. If using
    sip4, then we copy values inside the enum types as sip6 does.

    If recursive (the default), sub-modules are also scanned and modified.
    '''
    sip = sys.modules['sip']
    if sip.SIP_VERSION >= 0x060000:
        sip6_to_sip4(module, recursive=recursive)
    else:
        sip4_to_sip6_enums(module, recursive=recursive)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import os
import sys
import logging

from soma.qt_gui.qt_backend import QtGui, QtCore
from soma.qt_gui import qt_backend
from .File import FileControlWidget
from soma.utils.weak_proxy import get_ref
import traits.api as traits
import six


class DirectoryControlWidget(FileControlWidget):

    """ Control to enter a directory.
    """

    @staticmethod
    def is_valid(control_instance, *args, **kwargs):
        """ Method to check if the new control value is correct.

        If the new entered value is not correct, the backroung control color
        will be red.

        Parameters
        ----------
        control_instance: QWidget (mandatory)
            the control widget we want to validate

        Returns
        -------
        out: bool
            True if the control value is a file,
            False otherwise
        """
        # Get the current control palette
        control_palette = control_instance.path.palette()

        # Get the control current value
        control_value = control_instance.path.value()

        color = QtCore.Qt.white
        red = QtGui.QColor(255, 220, 220)
        yellow = QtGui.QColor(255, 255, 200)

        # If the control value contains a file, the control is valid and the
        # backgound color of the control is white
        is_valid = False
        if control_value is traits.Undefined:
            # Undefined is an exception: allow to reset it (File instances,
            # even mandatory, are initialized with Undefined value)
            is_valid = True
            if not control_instance.optional:
                color = red
        else:

            if os.path.isdir(control_value) \
                    or (control_instance.output and control_value != "") \
                    or (control_instance.trait.handler.exists is False
                        and control_value != ""):
                is_valid = True

            # If the control value is optional, the control is valid and the
            # backgound color of the control is yellow
            elif control_instance.optional is True and control_value == "":
                color = yellow
                is_valid = True

            # If the control value is empty, the control is not valid and the
            # backgound color of the control is red
            else:
                color = red

        # Set the new palette to the control instance
        control_palette.setColor(control_instance.path.backgroundRole(), color)
        control_instance.path.setPalette(control_palette)

        return is_valid

    #
    # Callbacks
    #

    @staticmethod
    def onBrowseClicked(control_instance):
        """ Browse the file system and update the control instance accordingly.

        If a valid direcorty has already been entered the dialogue will
        automatically point to this folder, otherwise the current working
        directory is used.

        Parameters
        ----------
        control_instance: QWidget (mandatory)
            the directory widget item
        """
        # Get the current directory
        current_control_value = os.path.join(os.getcwd(), os.pardir)
        if DirectoryControlWidget.is_valid(control_instance):
            current_control_value = six.text_type(control_instance.path.text())

        # Create a dialog to select a directory
        folder = qt_backend.getExistingDirectory(
            get_ref(control_instance), "Open directory", current_control_value,
            QtGui.QFileDialog.ShowDirsOnly
                | QtGui.QFileDialog.DontUseNativeDialog)

        # Set the selected directory to the path sub control
        control_instance.path.setText(six.text_type(folder))



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.


'''
Base classes for writing various minf formats (XML, HDF5, Python's pickle, etc.)

* author: Yann Cointepas
* organization: `NeuroSpin <http://www.neurospin.org>`_
* license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
'''
from __future__ import absolute_import
__docformat__ = "restructuredtext en"

import six
from soma.translation import translate as _


#------------------------------------------------------------------------------
class RegisterMinfWriterClass(type):

    '''
    RegisterMinfWriterClass is used as metaclass of L{MinfWriter} to automatically
    register all classes derived from L{MinfWriter}.
    '''
    def __init__(cls, name, bases, dict):
        # why test hasattr(cls, name) ?
        # on Ubuntu 12.04 the six.with_metaclass() function may trigger this
        # constructor on a "NewBase" type which doesn't have the name attribute
        if hasattr(cls, 'name') and cls.name is not None:
            MinfWriter._allWriterClasses[cls.name] = cls


#------------------------------------------------------------------------------
class MinfWriter(six.with_metaclass(RegisterMinfWriterClass, object)):

    '''
    Class derived from MinfWriter are responsible of writing a specific format of
    minf file. This version only support XML format but other formats may
    be added later (such as HDF5).
    All classes derived from L{MinfWriter} must define a L{name} class attribute
    the following methods:
      - L{__init__} to construct writer instances.
      - L{write} to write objects in minf file.
      - L{close} to terminate writing.
    '''

    #: all classes derived from L{MinfWriter} are automatically stored in that
    #: dictionary (keys are formats name and values are class objects).
    _allWriterClasses = {}

    #: class derived from L{MinfWriter} must set a format name in this attribute.
    name = None

    def __init__(self, file, reducer):
        '''
        Constructor of classes derived from L{MinfWriter} must be callable with two
        parameters.
        @param file: file object (opened for writing) where the minf file is
          written.
        @type  file: any object respecting Python file object API
        @param reducer: name of the reducer to use (see L{soma.minf.tree} for
          more information about reducers).
        @type  reducer: string
        '''

    def write(self, value):
        '''
        Write an object into the minf file.
        @param value: any value that can be written in this minf file.
        '''

    def close(self):
        '''
        Close the writer, further calls to L{write} method will lead to an error.
        The underlying file is NOT closed.
        '''

    def createWriter(destFile, format, reducer):
        '''
        This static method create a L{MinfWriter} instance by looking for a
        registered L{MinfWriter} derived classe named C{format}. Parameters
        C{destFile} and C{reducer} are passed to the derived class constructor.
        @param format: name of the minf format.
        @type  format: string
        @returns: L{MinfWriter} derived class instance.
        @param file: file name or file object (opened for writing) where the minf
          file is written. If it is a file name, it is opened with
          C{open( destFile, 'wb' )}.
        @type  file: string or any object respecting Python file object API
        @param reducer: name of the reducer to use (see L{soma.minf.tree} for
          more information about reducers).
        @type  reducer: string
        '''
        writer = MinfWriter._allWriterClasses.get(format)
        if writer is None:
            raise ValueError(
                _('No minf writer for format "%(format)s", possible formats are: %(possible)s')
                %
                {'format': format,
                 'possible': ', '.join(['"' + i + '"'
                                        for i in
                                        MinfWriter._allWriterClasses])})
        if not hasattr(destFile, 'write'):
            destFile = open(destFile, 'w')
        return writer(destFile, reducer, )
    createWriter = staticmethod(createWriter)



# -*- coding: utf-8 -*-
#
# SOMA - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
#

'''Base classes for the management of association between Python objects and
factories (i.e. a function that create something for an object) taking into
account classes hierachy but without storing anything in the objects classes.
'''

from __future__ import absolute_import

from weakref import WeakKeyDictionary


class MetaFactories(type):

    '''This is the metaclass of Factories.
    It intercepts Factories class creation to make sure that they have
    a _global_factories attibute.'''

    def __new__(mcs, name, bases, dictionary):
        if '_global_factories' not in dictionary:
            dictionary['_global_factories'] = {}
        return type.__new__(mcs, name, bases, dictionary)


class Factories(object):

    '''
    This is the base class for managing association between any Python object
    and a factory. For instance, WidgetFactories is derived from Factories and
    used to register all factories allowing to create a graphical widget for a
    Python object.

    There are two levels of associations between Python objects and factories
    (i.e. two dictionaries). The global factories are associated at the class
    level (for classes derived from Factories) with the register_global_factory
    method. The global factories are shared by all instances. On the other hand,
    if one wants to customize some factories for a specific context, it can
    create one Factories instance per context and use register_factory method
    to create an associtation at the instance level.
    '''
    __metaclass__ = MetaFactories
    _global_factories = {}

    def __init__(self):
        super(Factories, self).__init__()
        self._factories = WeakKeyDictionary()

    @classmethod
    def register_global_factory(cls, klass, factory):
        '''
        Create an association between a class and a global factory.
        '''
        cls._global_factories[klass] = factory

    def get_global_factory(self, klass):
        '''
        Retrieve the global factory associated to a class or an instance. Only
        direct association is used. In order to take into account class hierarchy,
        one must use get_factory method.
        '''
        return self._global_factories.get(klass)

    def register_factory(self, class_or_instance, factory):
        '''
        Create an association between a class (or an instance) and a factory.
        '''
        self._factories[class_or_instance] = factory

    def get_factory(self, class_or_instance):
        '''
        Retrieve the factory associated to an object.
        First look into the object instance and then in the object class hierarchy.
        At each step a registered factory in the Factories instance is looked for.
        If there is none, self.get_global_factory is used.
        Returns None if no factory is found.
        '''
        for key in (class_or_instance, ) + class_or_instance.__class__.__mro__:
            factory = self._factories.get(key)
            if factory is not None:
                break
            factory = self.get_global_factory(key)
            if factory is not None:
                break
        return factory



# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import absolute_import
from soma.controller import Controller
import six

try:
    from traits.api import Undefined
except ImportError:
    class Undefined(object):
        pass

def to_json(value):
    '''
    Convert value to an object which will mark some types through JSON
    serialization. Typically, tuples will be replaced with lists which firsst
    element is '<tuple>', Undefined with ['<undefined'], sets with ['<set>,
    items], etc.

    "Decding" can be done using :func:`from_json`
    '''
    if isinstance(value, tuple):
        value = ['<tuple>'] + [to_json(x) for x in value]
    if isinstance(value, set):
        value = ['<set>'] + [to_json(x) for x in value]
    elif isinstance(value, list):
        value = [to_json(x) for x in value]
    elif isinstance(value, Controller):
        value = to_json(value.export_to_dict())
    elif getattr(value, 'items', None):
        # (hasattr may answer True for HasTraits)
        new_value = {}
        for key, item in six.iteritems(value):
            new_value[key] = to_json(item)
        value = new_value
    elif value is Undefined:
        value = ['<undefined>']
    return value


def from_json(value):
    '''
    Reverse of :func:`to_json`

    Convert value from an object which matches JSON serialization, containing
    "code" for some types. Typically, tuples, sets, Undefined, etc.
    '''
    if hasattr(value, 'items'):
        new_value = type(value)()
        for key, item in six.iteritems(value):
            new_value[key] = from_json(item)
        return new_value
    if not isinstance(value, list):
        return value
    if len(value) < 1:
        return value
    code = value[0]
    if code == '<tuple>':
        return tuple([from_json(x) for x in value[1:]])
    elif code == '<undefined>':
        return Undefined
    elif code == '<set>':
        return set([from_json(x) for x in value[1:]])
    return [from_json(x) for x in value]
